from __future__ import annotations

import abc
import logging
import os
import pickle
import re
import subprocess
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Optional, cast
from urllib.error import HTTPError

from ray.air.integrations.wandb import WandbLoggerCallback, _clean_log, _QueueItem, _WandbLoggingActor
from ray.tune.utils import flatten_dict

from ray_utilities import RUN_ID
from ray_utilities.callbacks.tuner.new_style_logger_callback import LogMetricsDictT, NewStyleLoggerCallback
from ray_utilities.callbacks.tuner.track_forked_trials import TrackForkedTrialsMixin
from ray_utilities.comet import _LOGGER
from ray_utilities.constants import DEFAULT_VIDEO_DICT_KEYS, FORK_FROM
from ray_utilities.misc import extract_trial_id_from_checkpoint, make_experiment_key, parse_fork_from

if TYPE_CHECKING:
    from ray_utilities.typing.metrics import AnyFlatLogMetricsDict
    from ray_utilities.typing import ForkFromData

from ._log_result_grouping import non_metric_results
from ._save_video_callback import SaveVideoFirstCallback

if TYPE_CHECKING:
    from ray.tune.experiment import Trial
    from wandb.sdk.interface.interface import PolicyName

    from ray_utilities.typing.metrics import (
        VideoMetricsDict,
        _LogMetricsEvalEnvRunnersResultsDict,
    )

try:
    from wandb import Artifact, Video
except ImportError:
    pass  # wandb not installed
else:
    from ray.air.integrations import wandb as ray_wandb
    from wandb.errors import CommError

    def _is_allowed_type_patch(obj):
        """Return True if type is allowed for logging to wandb"""
        if _original_is_allowed_type(obj):
            return True
        return isinstance(obj, (FutureFile, FutureArtifact))

    _original_is_allowed_type = ray_wandb._is_allowed_type
    ray_wandb._is_allowed_type = _is_allowed_type_patch

_logger = logging.getLogger(__name__)


class _WandbLoggingActorWithArtifactSupport(_WandbLoggingActor):
    def run(self, retries=0):
        fork_from = self.kwargs.get("fork_from", None) is not None
        if fork_from:
            # Write info about forked trials, to know in which order to upload trials
            info_file = Path(self._logdir) / "wandb_fork_from.txt"
            if not info_file.exists():
                # write header
                info_file.write_text("trial_id, parent_id, parent_step, step_metric\n")
            fork_data = parse_fork_from(self.kwargs["fork_from"])
            with info_file.open("a") as f:
                if fork_data is not None:
                    parent_id, parent_step = fork_data
                    f.write(f"{self.kwargs['id']}, {parent_id}, {parent_step}, _step\n")
                else:
                    _logger.error("Could not parse fork_from: %s", self.kwargs["fork_from"])
                    f.write(f"{self.kwargs['id']}, {self.kwargs['fork_from']}\n")
        try:
            return super().run()
        except CommError as e:  # pyright: ignore[reportPossiblyUnboundVariable]
            # NOTE: its possible that wandb is stuck because of async logging and we never reach here :/
            # breakpoint()
            online = self.kwargs.get("mode", "online") == "online"
            if "fromStep is greater than the run's last step" in str(e):
                # This can happen if the parent run is not yet fully synced.
                if not fork_from:
                    raise  # should only happen on forks
                if retries >= 5:
                    _logger.error(
                        "WandB communication error. online mode: %s, fork_from: %s - Error: %s", online, fork_from, e
                    )
                    if not online:
                        raise
                    _logger.warning("Switching to offline mode")
                    self.kwargs["mode"] = "offline"
                    return super().run()
                _logger.warning("WandB communication error, retrying after 10s: %s", e)
                time.sleep(10)
                return self.run(retries=retries + 1)
            if not online:
                _logger.exception("WandB communication error in offline mode. Cannot recover.")
                raise
            if fork_from:
                _logger.error("WandB communication error when using fork_from")
            _logger.exception("WandB communication error. Trying to switch to offline mode.")
            self.kwargs["mode"] = "offline"
            super().run()
            # TODO: communicate to later upload offline run

    def _handle_result(self, result: dict) -> tuple[dict, dict]:
        config_update = result.get("config", {}).copy()
        log = {}
        flat_result: AnyFlatLogMetricsDict | dict[str, Any] = flatten_dict(result, delimiter="/")

        for k, v in flat_result.items():
            if any(k.startswith(item + "/") or k == item for item in self._exclude):
                continue
            if any(k.startswith(item + "/") or k == item for item in self._to_config):
                config_update[k] = v
            elif isinstance(v, FutureFile):
                try:
                    self._wandb.save(v.global_str, base_path=v.base_path)
                except (HTTPError, Exception) as e:  # noqa: BLE001
                    _logger.error("Failed to log artifact: %s", e)
            elif isinstance(v, FutureArtifact):
                # not serializable
                artifact = Artifact(  # pyright: ignore[reportPossiblyUnboundVariable]
                    name=v.name,
                    type=v.type,
                    description=v.description,
                    metadata=v.metadata,
                    incremental=v.incremental,
                    **v.kwargs,
                )
                for file_dict in v._added_files:
                    artifact.add_file(**file_dict)
                for dir_dict in v._added_dirs:
                    artifact.add_dir(**dir_dict)
                for ref_dict in v._added_references:
                    artifact.add_reference(**ref_dict)
                try:
                    self._wandb.log_artifact(artifact)
                except (HTTPError, Exception) as e:
                    _logger.error("Failed to log artifact: %s", e)
            elif not _is_allowed_type_patch(v):
                continue
            else:
                log[k] = v

        config_update.pop("callbacks", None)  # Remove callbacks
        return log, config_update


class AdvWandbLoggerCallback(
    NewStyleLoggerCallback, SaveVideoFirstCallback, TrackForkedTrialsMixin, WandbLoggerCallback
):
    AUTO_CONFIG_KEYS: ClassVar[list[str]] = list(
        {
            *WandbLoggerCallback.AUTO_CONFIG_KEYS,
            *non_metric_results,
        }
    )

    _logger_actor_cls = _WandbLoggingActorWithArtifactSupport

    _logged_architectures: set[Trial]

    def __init__(
        self,
        *,
        project: Optional[str] = None,
        group: Optional[str] = None,
        excludes: Optional[list[str]] = None,
        upload_checkpoints: bool = False,
        video_kwargs: Optional[dict] = None,
        image_kwargs: Optional[dict] = None,
        upload_offline_experiments: bool = False,
        **kwargs,
    ):
        """For ``kwargs`` see :class:`ray.air.integrations.wandb.WandbLoggerCallback`"""
        kwargs.update(
            {
                "project": project,
                "group": group,
                "excludes": excludes or [],
                "upload_checkpoints": upload_checkpoints,
                "video_kwargs": video_kwargs,
                "image_kwargs": image_kwargs,
            }
        )
        super().__init__(**kwargs)
        self._trials_created = 0
        self._trials_started = 0
        """A Trial can be started multiple times due to restore."""
        self._logged_architectures = set()
        self.upload_offline_experiments = upload_offline_experiments
        """If True, offline experiments will be uploaded on trial completion."""

    def on_trial_start(self, iteration: int, trials: list["Trial"], trial: "Trial", **info):
        super().on_trial_start(iteration, trials, trial, **info)
        _logger.debug("Trials created: %d, re-started: %d", self._trials_created, self._trials_started)
        self._trials = trials  # keep them in case of a failure to access paths.

    def log_trial_start(self, trial: "Trial"):
        # breakpoint()
        config = trial.config.copy()

        config.pop("callbacks", None)  # Remove callbacks
        config.pop("log_level", None)

        exclude_results = self._exclude_results.copy()

        # Additional excludes
        exclude_results += self.excludes

        # Log config keys on each result?
        if not self.log_config:
            exclude_results += ["config"]

        # Fill trial ID and name
        trial_id = trial.trial_id if trial else None
        trial_name = str(trial) if trial else None

        # Project name for Wandb
        wandb_project = self.project

        # Grouping
        wandb_group = self.group or trial.experiment_dir_name if trial else None

        # remove unpickleable items!
        config: dict[str, Any] = _clean_log(config)  # pyright: ignore[reportAssignmentType]
        config = {key: value for key, value in config.items() if key not in self.excludes}
        config["run_id"] = RUN_ID
        # replace potential _ in trial_id
        # --- New Code --- : Remove nested keys
        for nested_key in filter(lambda x: "/" in x, self.excludes):
            key, sub_key = nested_key.split("/")
            if key in config:
                config[key].pop(sub_key, None)
        fork_from = fork_id = fork_iteration = None  # new run
        if "cli_args" in config:
            assert "num_jobs" not in config["cli_args"]
            assert "test" not in config["cli_args"]
            if trial.config["cli_args"].get("from_checkpoint"):
                fork_id = extract_trial_id_from_checkpoint(trial.config["cli_args"]["from_checkpoint"])
                # get id of run
                if fork_id is None:
                    _logger.error(
                        "Cannot extract trial id from checkpoint name: %s. "
                        "Make sure that it has to format id=<part1>_<sample_number>",
                        trial.config["cli_args"]["from_checkpoint"],
                    )
                else:
                    # Need to change to format '<run>?<metric>=<numeric_value>'
                    # Where metric="_step"; open state pickle to get iteration
                    ckpt_dir = Path(trial.config["cli_args"]["from_checkpoint"])
                    state = None
                    if (state_file := ckpt_dir / "state.pkl").exists():
                        with open(state_file, "rb") as f:
                            state = pickle.load(f)
                    elif (ckpt_dir / "_dict_checkpoint.pkl").exists():
                        with open(ckpt_dir / "_dict_checkpoint.pkl", "rb") as f:
                            state = pickle.load(f)["state"]
                    if state is None:
                        _logger.error(
                            "Could not find state.pkl or _dict_checkpoint.pkl in the checkpoint path. "
                            "Cannot use fork_from with wandb"
                        )
                    else:
                        iteration = state["trainable"]["iteration"]
                        fork_from = f"{fork_id}?_step={iteration}"
                fork_iteration = None  # NOTE: Cannot fork twice in same run; would need Checkpoint to determine step
        # we let take FORK_FROM a higher priority
        if FORK_FROM in trial.config:
            fork_data = cast("ForkFromData", trial.config[FORK_FROM])
            fork_id = fork_data["parent_id"]
            fork_iteration = fork_data["parent_training_iteration"]
            fork_from = f"{fork_id}?_step={fork_iteration}"
            # We should not have multiple ?_step= in the id
            trial_id = self.get_forked_trial_id(trial)
            assert trial_id is not None, "Expected trial_id to be set on super for forked trial."
            trial_name = self.make_forked_trial_name(trial, fork_data)
            # Set experiment key using dict-based fork data
            config.setdefault("experiment_key", make_experiment_key(trial, fork_data))
        else:
            # No fork info present in config; use non-fork key
            config.setdefault("experiment_key", make_experiment_key(trial))

        # TODO: trial_id of a forked trial might be very long

        # Test for invalid chars
        assert not trial_id or all(c not in trial_id for c in r"/ \ # ? % :"), f"Invalid character in: {trial_id}"
        assert fork_from is None or fork_from.count("?_step=") == 1, fork_from
        # FIXME: fork_from is not stable on WandB when there is no longer wait time, wandb.init likely raises error
        # NOTE: We never want FORK_FROM to be in the trials.config by default.
        # --- End New Code
        wandb_init_kwargs = {
            "id": trial_id,  # change if forked? e.g. + forked_from
            "name": trial_name,
            "reinit": "default",  # bool is deprecated
            "allow_val_change": True,
            "group": wandb_group,
            "project": wandb_project,
            "config": config,
            # possibly fork / resume
            "fork_from": fork_from,
        }
        wandb_init_kwargs.update(self.kwargs)
        if fork_from:
            wandb_init_kwargs.setdefault("tags", []).append("forked")

        if trial not in self._trial_logging_actors:
            self._trials_created += 1
        if fork_from and trial in self._trial_logging_futures:
            assert self.is_trial_forked(trial), "Expected trial to be tracked as forked trial."
            self._restart_logging_actor(trial, **wandb_init_kwargs)
        else:
            # can be forked from a checkpoint, if not stopped does not start a new
            self._start_logging_actor(trial, exclude_results, **wandb_init_kwargs)
        self._trials_started += 1

    def _restart_logging_actor(self, trial: "Trial", **wandb_init_kwargs):
        """Ends the current logging actor and starts a new one. Useful for resuming with a new ID / settings."""
        self.log_trial_end(trial, failed=False)
        _logger.info("Restarting WandB logging actor for trial %s", trial.trial_id)
        # Wait a bit before starting the next one
        self._cleanup_logging_actors(timeout=5, kill_on_timeout=False)
        # Clean queue and futures else a new one will not be created
        self._trial_queues.pop(trial, None)
        self._trial_logging_futures.pop(trial, None)
        self._trial_logging_actors.pop(trial, None)
        self._start_logging_actor(trial, self._exclude_results, **wandb_init_kwargs)

    @staticmethod
    def preprocess_videos(metrics: LogMetricsDictT) -> LogMetricsDictT:
        did_copy = False
        for keys in DEFAULT_VIDEO_DICT_KEYS:
            subdir = metrics
            for key in keys[:-1]:
                if key not in subdir:
                    break
                subdir = subdir[key]  # pyright: ignore[reportGeneralTypeIssues, reportTypedDictNotRequiredAccess]
            else:
                # Perform a selective deep copy on the modified items
                subdir = cast("dict[str, VideoMetricsDict]", subdir)
                if keys[-1] in subdir and "video_path" in subdir[keys[-1]]:
                    if not did_copy:
                        metrics = metrics.copy()  # pyright: ignore[reportAssignmentType]
                        did_copy = True
                    parent_dir = metrics
                    for key in keys[:-1]:
                        parent_dir[key] = parent_dir[key].copy()  # pyright: ignore[reportGeneralTypeIssues, reportTypedDictNotRequiredAccess]  # fmt: skip
                        parent_dir = parent_dir[key]  # pyright: ignore[reportGeneralTypeIssues, reportTypedDictNotRequiredAccess]  # fmt: skip
                    parent_dir = cast("_LogMetricsEvalEnvRunnersResultsDict", parent_dir)
                    parent_dir[keys[-1]] = video_dict = cast("VideoMetricsDict", parent_dir[keys[-1]]).copy()  # pyright: ignore[reportTypedDictNotRequiredAccess]  # fmt: skip
                    # IMPORTANT use absolute path as local path is a ray session!
                    video_dict["video"] = Video(  # pyright: ignore[reportPossiblyUnboundVariable]
                        os.path.abspath(video_dict.pop("video_path")), format="mp4"
                    )

        return metrics  # type: ignore[return-value]

    def on_trial_complete(self, iteration: int, trials: list["Trial"], trial: "Trial", **info):
        """Called when a trial has completed. Triggers sync for offline runs."""
        # Call parent method to handle normal trial completion
        super().on_trial_complete(iteration, trials, trial, **info)

        # TODO: Also sync if the trial will be perturbed; but it will not reach on_trial_complete!
        # Furthermore on_trial_complete there will be multiple folders.

        # If we are in offline mode, try to sync this trial's run immediately
        if "offline" in self.kwargs.get("mode", "") and self.upload_offline_experiments:
            # Wandb dir is likely not yet saved by actor, wait for it, super does not wait that long.
            self._cleanup_logging_actors(timeout=120, kill_on_timeout=False)
            _LOGGER.info("Syncing offline WandB run for trial %s", trial.trial_id)
            # NOTE: Actor should have synced everything at this point
            self._sync_offline_run_if_available(trial)

    def _sync_offline_run_if_available(self, trial: "Trial"):
        """Sync offline WandB run for the given trial if it exists."""
        try:
            # Look for offline runs that might belong to this trial
            assert trial.local_path
            wandb_dir = Path(trial.local_path) / "wandb"  # might not be accessible
            wait = 5
            while not wandb_dir.exists() and wait < 30:
                _logger.debug("WandB directory does not exist yet, waiting %s/30s: %s", wait, wandb_dir)
                time.sleep(5)  # wait for possible sync
                wait += 5
            if not wandb_dir.exists() and trial.path is not None:
                _logger.debug("WandB directory does not exist on Tuner system %s", wandb_dir)
                # Trigger a sync from local -> remote
                if trial.storage:
                    # local_experiment_path will always work but is overkill, try only wandb folder
                    sync_locations: list[tuple[str, str]] = [
                        (trial.local_experiment_path, trial.remote_experiment_path)
                    ]
                    sync_locations.insert(0, (wandb_dir.as_posix(), (Path(trial.path) / "wandb").as_posix()))
                    for local_path, remote_path in sync_locations:
                        try:
                            if trial.storage.syncer.sync_up(
                                local_path,
                                remote_path,
                            ):
                                trial.storage.syncer.wait()
                        except FileNotFoundError:  # noqa: PERF203
                            pass
                # Remote path
                wandb_dir = Path(trial.path) / "wandb"
                if not wandb_dir.exists():
                    _logger.debug("WandB directory does not exist: %s", wandb_dir)
                    return

            # Wandb file should be bound to the trial and not duplicated
            offline_runs = list(wandb_dir.glob("offline-run-*"))
            if len(offline_runs) > 1 and FORK_FROM not in trial.config:
                # This is normal when having a forked trial or it was forked in the past
                _logger.warning("Multiple wandb offline directories found in %s: %s", wandb_dir, offline_runs)

            if not offline_runs:
                _logger.error(
                    "No wandb offline experiments found to upload in %s: %s. ", wandb_dir, list(wandb_dir.glob("*"))
                )
                return
            # Sort by modification time and take the most recent

            # when not forked likely just one item
            # TODO: Save a file with commands to upload again in case a run fails!
            for run_dir in sorted(offline_runs, key=lambda p: p.stat().st_mtime, reverse=True):
                # Use wandb sync command to upload the offline run
                _logger.info("Attempting to sync offline WandB run: %s", run_dir)
                result = subprocess.run(
                    ["wandb", "sync", str(run_dir)],
                    check=False,
                    text=True,
                    timeout=600,  # timeout 10 minutes
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
                if result.returncode == 0 and "error" not in result.stdout.lower():
                    _logger.info("Successfully synced offline run for trial %s\n%s", trial.trial_id, result.stdout)
                elif "not found (<Response [404]>)" in result.stdout:
                    _logger.error(
                        "Could not sync run for trial %s "
                        "(Is it a forked_run? - The parent needs to be uploaded first): %s",
                        trial.trial_id,
                        result.stdout,
                    )
                else:
                    _logger.error("Error during syncing offline run for trial %s: %s", trial.trial_id, result.stdout)
                if result.returncode != 0 or result.stderr:
                    _logger.error("Failed to sync offline run for trial %s: %s", trial.trial_id, result.stderr)
                if len(offline_runs) > 1:
                    time.sleep(5)  # wait a bit between uploads

        except subprocess.TimeoutExpired:
            _logger.warning("Timeout while syncing offline run for trial %s", trial.trial_id)
        except (OSError, subprocess.SubprocessError) as e:
            _logger.warning("Failed to sync offline run for trial %s: %s", trial.trial_id, e)

    def log_trial_result(
        self,
        iteration: int,  # noqa: ARG002
        trial: "Trial",
        result,
    ):
        """Called each time a trial reports a result."""
        if trial not in self._trial_logging_actors:
            self.log_trial_start(trial)
            # log model config
        if trial not in self._logged_architectures and "model_architecture.json" in os.listdir(trial.path):
            if trial.path is not None:
                result = result.copy()
                file_path = os.path.abspath(os.path.join(trial.path, "model_architecture.json"))
                artifact = FutureFile(file_path, Path(file_path).parent, policy="live")
                result["model_architecture"] = artifact  # pyright: ignore[reportGeneralTypeIssues]
                self._logged_architectures.add(trial)
                _LOGGER.debug("Storing future Artifact %s", artifact.to_dict())
            else:
                _LOGGER.error("Cannot save model_architecture as trial.path is None")

        result_clean = _clean_log(self.preprocess_videos(result))
        if not self.log_config:
            # Config will be logged once log_trial_start
            result_clean.pop("config", None)  # type: ignore
        self._trial_queues[trial].put((_QueueItem.RESULT, result_clean))


class _WandbFuture(abc.ABC):
    @abc.abstractmethod
    def json_encode(self) -> dict[str, Any]: ...

    def to_dict(self):
        return self.json_encode()


class FutureFile(_WandbFuture):
    """A file to be logged to WandB for this run, has to be compatible with :meth:`wandb.save`."""

    def __init__(
        self,
        glob_str: str | os.PathLike,
        base_path: str | os.PathLike | None = None,
        policy: PolicyName = "live",
    ) -> None:
        self.global_str = glob_str
        self.base_path = base_path
        self.policy = policy

    def json_encode(self) -> dict[str, Any]:
        return {
            "glob_str": self.global_str,
            "base_path": self.base_path,
            "policy": self.policy,
        }


class FutureArtifact(_WandbFuture):
    def __init__(
        self,
        name: str,
        type: str,
        description: str | None = None,
        metadata: dict[str, Any] | None = None,
        *,
        incremental: bool = False,
        **kwargs,
    ):
        if not re.match(r"^[a-zA-Z0-9_\-.]+$", name):
            raise ValueError(
                f"Artifact name may only contain alphanumeric characters, dashes, "
                f"underscores, and dots. Invalid name: {name}"
            )
        self.name = name
        self.type = type
        self.description = description
        self.metadata = metadata
        self.incremental = incremental
        self.kwargs = kwargs
        self._added_dirs = []
        self._added_files = []
        self._added_references = []

    def add_reference(self, uri: Any | str, name: str | None = None, **kwargs) -> None:
        self._added_references.append({"uri": uri, "name": name, **kwargs})

    def add_file(
        self,
        local_path: str,
        name: str | None = None,
        *,
        is_tmp: bool | None = False,
        overwrite: bool = False,
        **kwargs,
    ) -> None:
        self._added_files.append(
            {
                "local_path": local_path,
                "name": name,
                "is_tmp": is_tmp,
                "overwrite": overwrite,
                **kwargs,
            }
        )

    def add_dir(
        self,
        local_path: str,
        name: str | None = None,
        **kwargs,
    ) -> None:
        self._added_dirs.append(
            {
                "local_path": local_path,
                "name": name,
                **kwargs,
            }
        )

    def json_encode(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "type": self.type,
            "description": self.description,
            "metadata": self.metadata,
            "incremental": self.incremental,
            "kwargs": self.kwargs,
            "added_dirs": self._added_dirs,
            "added_files": self._added_files,
            "added_references": self._added_references,
        }

    def to_dict(self) -> dict[str, Any]:
        return self.json_encode()



"""Extension mixins for dynamic configuration in Ray RLlib experiment setups.

This module provides mixin classes that extend the base experiment setup functionality
with dynamic configuration capabilities. These mixins can be composed with base setup
classes to add features like dynamic buffer sizing, batch size adjustment, and
adaptive evaluation intervals during training.

The mixins are designed to work together and with the base
:class:`~ray_utilities.setup.experiment_base.ExperimentSetupBase` class through
multiple inheritance, providing modular functionality that can be mixed and matched
based on experiment requirements.

Key Components:
    - :class:`SetupWithDynamicBuffer`: Dynamic experience buffer sizing
    - :class:`SetupWithDynamicBatchSize`: Dynamic batch size and gradient accumulation
    - :class:`SetupForDynamicTuning`: Base class for dynamic configuration mixins

These extensions integrate with Ray Tune's parameter spaces and RLlib callbacks
to provide adaptive behavior during training and hyperparameter optimization.
"""

from __future__ import annotations

import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar

from ray import tune

from ray_utilities.callbacks.algorithm.dynamic_batch_size import DynamicGradientAccumulation
from ray_utilities.callbacks.algorithm.dynamic_buffer_callback import DynamicBufferUpdate
from ray_utilities.callbacks.algorithm.dynamic_evaluation_callback import add_dynamic_eval_callback_if_missing
from ray_utilities.setup.experiment_base import AlgorithmType_co, ConfigType_co, ExperimentSetupBase, ParserType_co

if TYPE_CHECKING:
    from ray.rllib.callbacks.callbacks import RLlibCallback

    from ray_utilities.typing import ParameterSpace

_logger = logging.getLogger(__name__)


class SetupForDynamicTuning(ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]):
    ...
    # base class that can be expanded in the future


class SetupWithDynamicBuffer(SetupForDynamicTuning[ParserType_co, ConfigType_co, AlgorithmType_co]):
    """Mixin class for dynamic experience buffer sizing in RLlib experiments.

    This mixin adds the capability to dynamically adjust the experience buffer
    (rollout) size during training or hyperparameter optimization. It provides
    both predefined parameter spaces for tuning and automatic callback integration
    for adaptive buffer management.

    The mixin automatically adds the :class:`~ray_utilities.callbacks.algorithm.dynamic_buffer_callback.DynamicBufferUpdate`
    callback when dynamic buffer sizing is enabled, and includes evaluation
    interval adjustment to work properly with the dynamic buffer updates.

    Features:
        - Predefined rollout size parameter space for Ray Tune optimization
        - Automatic callback registration for dynamic buffer updates
        - Integration with dynamic evaluation intervals
        - Supports both grid search and adaptive parameter selection

    Class Attributes:
        rollout_size_sample_space: Ray Tune parameter space with common rollout sizes
            ranging from 32 to 8192 steps, suitable for grid search optimization.

    Note:
        This mixin should be used before other setups that add
        :class:`~ray_utilities.callbacks.algorithm.dynamic_evaluation_callback.DynamicEvalInterval`
        to avoid error logs about missing configuration keys.

    Example:
        class MySetup(SetupWithDynamicBuffer, ExperimentSetupBase):
            config_class = PPOConfig
            algo_class = PPO

    See Also:
        :class:`~ray_utilities.callbacks.algorithm.dynamic_buffer_callback.DynamicBufferUpdate`: Buffer update callback
        :class:`~ray_utilities.callbacks.algorithm.dynamic_evaluation_callback.DynamicEvalInterval`: Evaluation callback
        :class:`SetupWithDynamicBatchSize`: Companion mixin for batch size dynamics
    """

    rollout_size_sample_space: ClassVar[ParameterSpace[int]] = tune.grid_search(
        [32, 64, 128, 256, 512, 1024, 2048, 3072, 4096, 2048 * 3, 8192]  # 4096 * 3, 16384]
    )

    @classmethod
    def _get_callbacks_from_args(cls, args) -> list[type[RLlibCallback]]:
        # Can call the parent; might be None for ExperimentSetupBase
        callbacks = super()._get_callbacks_from_args(args)
        if args.dynamic_buffer:
            callbacks.append(DynamicBufferUpdate)
            add_dynamic_eval_callback_if_missing(callbacks)
        return callbacks

    @classmethod
    def _create_dynamic_buffer_params(cls):
        return deepcopy(cls.rollout_size_sample_space)

    def create_param_space(self) -> dict[str, Any]:
        self._set_dynamic_parameters_to_tune()  # sets _dynamic_parameters_to_tune
        if not self.args.tune or not (
            (add_all := "all" in self._dynamic_parameters_to_tune) or "rollout_size" in self._dynamic_parameters_to_tune
        ):
            return super().create_param_space()
        if not add_all:
            self._dynamic_parameters_to_tune.remove(
                "rollout_size"
            )  # remove before calling super().create_param_space()
        param_space = super().create_param_space()
        # TODO: # FIXME "rollout_size" is not used anywhere
        # however train_batch_size_per_learner is used with the DynamicBatchSize Setup
        # which uses in ints dynamic variant gradient accumulation.
        param_space["rollout_size"] = self._create_dynamic_buffer_params()
        return param_space


class SetupWithDynamicBatchSize(SetupForDynamicTuning[ParserType_co, ConfigType_co, AlgorithmType_co]):
    """Mixin class for dynamic batch size adjustment through gradient
    accumulation.

    This mixin enables dynamic batch size control during training by
    utilizing gradient accumulation rather than directly modifying the
    ``train_batch_size_per_learner``. It integrates the
    :class:`~ray_utilities.callbacks.algorithm.dynamic_batch_size.DynamicGradientAccumulation`
    callback when dynamic batch sizing is enabled so the effective batch
    size can be adjusted during training or tuning.

    Features:
        - Dynamic batch size control via gradient accumulation
        - Predefined batch size parameter space for Ray Tune optimization
        - Automatic callback registration for gradient accumulation
        - Integration with dynamic evaluation intervals

    Note:
        Use :class:`SetupWithDynamicBuffer` for direct tuning of rollout
        sizes. This mixin controls effective batch size via gradient
        accumulation.

    Warning:
        The ``batch_size`` tuning values refer to effective batch sizes
        achieved via gradient accumulation, not direct accumulation
        multipliers.

    Examples:
        .. code-block:: python

            class MySetup(SetupWithDynamicBatchSize, ExperimentSetupBase):
                config_class = PPOConfig
                algo_class = PPO

                def create_config(self, args):
                    config = super().create_config(args)
                    if args.dynamic_batch:
                        # Gradient accumulation will be handled automatically
                        pass
                    return config

    See Also:
        :class:`~ray_utilities.callbacks.algorithm.dynamic_batch_size.DynamicGradientAccumulation`:
            Gradient accumulation callback
        :class:`SetupWithDynamicBuffer`:
            Companion mixin for buffer size dynamics
        :class:`~ray_utilities.callbacks.algorithm.dynamic_evaluation_callback.DynamicEvalInterval`:
            Evaluation callback
    """

    batch_size_sample_space: ClassVar[ParameterSpace[int]] = tune.grid_search(
        [32, 64, 128, 256, 512, 1024, 2048, 3072, 4096, 2048 * 3, 8192]  # 4096 * 3, 16384]
    )
    """
    Tune parameter space with batch sizes from 32 to 16384.
    """

    @classmethod
    def _get_callbacks_from_args(cls, args) -> list[type[RLlibCallback]]:
        # When used as a mixin, can call the parent; might be None for ExperimentSetupBase
        callbacks = super()._get_callbacks_from_args(args)
        if args.dynamic_batch:
            callbacks.append(DynamicGradientAccumulation)
            add_dynamic_eval_callback_if_missing(callbacks)
        return callbacks

    @classmethod
    def _create_dynamic_batch_size_params(cls):
        """Create parameter space for dynamic batch size tuning.

        Returns a deep copy of the class's batch size sample space for use in
        hyperparameter optimization. The returned parameter space contains
        batch size values that will be achieved through gradient accumulation.

        Returns:
            A Ray Tune parameter space containing batch size values for optimization.

        Note:
            The returned parameters represent effective batch sizes achieved through
            gradient accumulation, not the gradient accumulation multiplier values directly.

        Warning:
            This method adds ``batch_size`` parameters to the tuning space, not
            values for direct gradient accumulation control.
        """
        # TODO: control this somehow via args
        return deepcopy(cls.batch_size_sample_space)

    def create_param_space(self) -> dict[str, Any]:
        self._set_dynamic_parameters_to_tune()  # sets _dynamic_parameters_to_tune
        if not self.args.tune or not (
            (add_all := "all" in self._dynamic_parameters_to_tune) or "batch_size" in self._dynamic_parameters_to_tune
        ):
            return super().create_param_space()
        if not add_all:
            self._dynamic_parameters_to_tune.remove("batch_size")  # remove before calling super().create_param_space()
        param_space = super().create_param_space()
        param_space["train_batch_size_per_learner"] = self._create_dynamic_batch_size_params()
        return param_space



# ruff: noqa: ARG002

from __future__ import annotations

from typing import Any

from ray.rllib.connectors.connector_v2 import ConnectorV2
from ray.rllib.connectors.env_to_module import NumpyToTensor


class DummyConnector(ConnectorV2):
    """Minimal pass-through connector for Ray RLlib data processing pipelines.

    This connector provides a no-op implementation that returns batch data unchanged.
    It's useful as a placeholder when a connector is required by the framework but
    no actual data processing is needed.

    The connector maintains full compatibility with the ConnectorV2 interface while
    performing zero transformations on the data, making it ideal for testing or
    scenarios where data should flow through unmodified.

    Example:
        >>> connector = DummyConnector()
        >>> result = connector(batch={"obs": observation_data})
        >>> # result == {"obs": observation_data}

    See Also:
        :class:`ray.rllib.connectors.connector_v2.ConnectorV2`: Base connector class
        :class:`DummyNumpyToTensor`: Dummy version of numpy-to-tensor conversion
    """

    def __call__(
        self,
        *,
        batch: dict[str, Any],
        **kwargs,
    ) -> Any:
        """Return the input batch unchanged.

        Args:
            batch: Dictionary containing batch data to process.
            **kwargs: Additional keyword arguments (ignored).

        Returns:
            The input batch without any modifications.
        """
        return batch


class DummyNumpyToTensor(DummyConnector, NumpyToTensor):
    """Dummy connector combining pass-through behavior with NumpyToTensor interface.

    This connector inherits from both :class:`DummyConnector` and
    :class:`ray.rllib.connectors.env_to_module.NumpyToTensor`, providing a no-op
    implementation when tensor conversion is expected but not needed.

    It's particularly useful when the framework expects a NumpyToTensor connector
    but the data is already in the appropriate format or conversion should be bypassed.

    See Also:
        :class:`DummyConnector`: Base dummy connector implementation
        :class:`ray.rllib.connectors.env_to_module.NumpyToTensor`: Tensor conversion connector
    """



from __future__ import annotations

from typing import TYPE_CHECKING

from ray.tune import Callback

from .adv_comet_callback import AdvCometLoggerCallback
from .adv_csv_callback import AdvCSVLoggerCallback
from .adv_json_logger_callback import AdvJsonLoggerCallback
from .adv_tbx_logger_callback import AdvTBXLoggerCallback
from .adv_wandb_callback import AdvWandbLoggerCallback
from .sync_config_files_callback import SyncConfigFilesCallback

if TYPE_CHECKING:
    from ray.tune.callback import Callback

__all__ = [
    "AdvCSVLoggerCallback",
    "AdvCometLoggerCallback",
    "AdvJsonLoggerCallback",
    "AdvTBXLoggerCallback",
    "AdvWandbLoggerCallback",
    "SyncConfigFilesCallback",
]


DEFAULT_TUNER_CALLBACKS_NO_RENDER: list[type["Callback"]] = []
"""
Default callbacks to use when neither needing render_mode nor advanced loggers.

Note:
    AdvCometLoggerCallback is not included
"""

DEFAULT_TUNER_CALLBACKS_RENDER: list[type["Callback"]] = [
    AdvJsonLoggerCallback,
    AdvTBXLoggerCallback,
    AdvCSVLoggerCallback,
]
"""Default callbacks to use when needing render_mode"""

DEFAULT_ADV_TUNER_CALLBACKS = DEFAULT_TUNER_CALLBACKS_RENDER.copy()
"""
List of advanced tuner callbacks to use if the advanced variants should be used.
Recommended when using schedulers working with :const:`FORK_FROM`.

A copy of :obj:`DEFAULT_TUNER_CALLBACKS_RENDER`.
"""


def create_tuner_callbacks(*, adv_loggers: bool) -> list["Callback"]:
    if adv_loggers:
        return [cb() for cb in DEFAULT_ADV_TUNER_CALLBACKS]
    return [cb() for cb in DEFAULT_TUNER_CALLBACKS_NO_RENDER]



"""Custom Ray RLlib connectors for data processing and debugging.

Provides custom connector implementations that extend Ray RLlib's connector framework
for processing data between environments, agents, and learners. Includes debugging
capabilities and specialized data processing connectors.

Key Components:
    - :class:`DebugConnector`: Debugging connector with logging capabilities
    - Data filtering and processing connectors
    - Integration with RLlib's ConnectorV2 framework
"""



"""Module for custom Ray Tune schedulers."""

from __future__ import annotations

import inspect
import logging
from ast import literal_eval
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, TypeAlias

from ray.tune.schedulers import PopulationBasedTraining
from tap import to_tap_class

from ray_utilities.constants import DEFAULT_EVAL_METRIC, EVAL_METRIC_RETURN_MEAN
from ray_utilities.tune.scheduler.top_pbt_scheduler import TopPBTTrialScheduler

if TYPE_CHECKING:
    from ray.tune.search.sample import Domain

logger = logging.getLogger(__name__)

_HPMutationsType: TypeAlias = dict[str, "dict[Any, Any] | list[Any] | tuple[Any, ...] | Callable[[], Any] | Domain"]


def _to_hyperparam_mutations(string: str) -> _HPMutationsType:
    """Convert a string representation of hyperparameter mutations to a dictionary."""
    try:
        return literal_eval(string)
    except (ValueError, SyntaxError) as e:
        raise ValueError(
            f"Invalid hyperparam_mutations format, must be parsable by :func:`ast.literal_eval`: {string}. "
            "Alternatively set it at runtime."
        ) from e


class PopulationBasedTrainingParser(to_tap_class(PopulationBasedTraining)):
    """
    Attributes:
        time_attr: The training result attr to use for comparing time.
            Note that you can pass in something non-temporal such as
            `training_iteration` as a measure of progress, the only requirement
            is that the attribute should increase monotonically.
        metric: The training result objective value attribute. Stopping
            procedures will use this attribute. If None but a mode was passed,
            the `ray.tune.result.DEFAULT_METRIC` will be used per default.
        mode: One of {min, max}. Determines whether objective is
            minimizing or maximizing the metric attribute.
        perturbation_interval: Models will be considered for
            perturbation at this interval of `time_attr`. Note that
            perturbation incurs checkpoint overhead, so you shouldn't set this
            to be too frequent.
        burn_in_period: Models will not be considered for
            perturbation before this interval of `time_attr` has passed. This
            guarantees that models are trained for at least a certain amount
            of time or timesteps before being perturbed.
        hyperparam_mutations: Hyperparams to mutate. The format is
            as follows: for each key, either a list, function,
            or a tune search space object (tune.loguniform, tune.uniform,
            etc.) can be provided. A list specifies an allowed set of
            categorical values. A function or tune search space object
            specifies the distribution of a continuous parameter. You must
            use tune.choice, tune.uniform, tune.loguniform, etc.. Arbitrary
            tune.sample_from objects are not supported.
            A key can also hold a dict for nested hyperparameters.
            You must specify at least one of `hyperparam_mutations` or
            `custom_explore_fn`.
            Tune will sample the search space provided by
            `hyperparam_mutations` for the initial hyperparameter values if the
            corresponding hyperparameters are not present in a trial's initial `config`.
        quantile_fraction: Parameters are transferred from the top
            `quantile_fraction` fraction of trials to the bottom
            `quantile_fraction` fraction. Needs to be between 0 and 0.5.
            Setting it to 0 essentially implies doing no exploitation at all.
        resample_probability: The probability of resampling from the
            original distribution when applying `hyperparam_mutations`. If not
            resampled, the value will be perturbed by a factor chosen from
            `perturbation_factors` if continuous, or changed to an adjacent value
            if discrete.
        perturbation_factors: Scaling factors to choose between when mutating
            a continuous hyperparameter.
        custom_explore_fn: You can also specify a custom exploration
            function. This function is invoked as `f(config)` after built-in
            perturbations from `hyperparam_mutations` are applied, and should
            return `config` updated as needed. You must specify at least one of
            `hyperparam_mutations` or `custom_explore_fn`.
        log_config: Whether to log the ray config of each model to
            local_dir at each exploit. Allows config schedule to be
            reconstructed.
        require_attrs: Whether to require time_attr and metric to appear
            in result for every iteration. If True, error will be raised
            if these values are not present in trial result.
        synch: If False, will use asynchronous implementation of
            PBT. Trial perturbations occur every perturbation_interval for each
            trial independently. If True, will use synchronous implementation
            of PBT. Perturbations will occur only after all trials are
            synced at the same time_attr every perturbation_interval.
            Defaults to False. See Appendix A.1 here
            https://arxiv.org/pdf/1711.09846.pdf.
    """

    mode: Literal["min", "max"] = "max"
    """One of {min, max}.Determines whether objective is minimizing or maximizing the metric attribute."""
    metric: str = EVAL_METRIC_RETURN_MEAN
    """The metric to be optimized as flat key, e.g. 'evaluation/env_runners/episode_return_mean'."""
    hyperparam_mutations: Optional[_HPMutationsType] = None
    require_attrs: bool = True
    synch: bool = True
    log_config: bool = True
    time_attr: str = "current_step"
    quantile_fraction: float = 0.1
    perturbation_interval: float = 100_000
    resample_probability: float = 1.0  # always resample

    # custom_args, remove before passing to PopulationBasedTraining
    use_native_pbt: bool = False
    """Do not use TopPBTTrialScheduler"""

    def set_hyperparam_mutations(self, mutations: _HPMutationsType | None) -> None:
        if mutations is None:
            self.hyperparam_mutations = None
            return
        mutations = mutations.copy()
        if "batch_size" in mutations:
            mutations["train_batch_size_per_learner"] = mutations.pop("batch_size")
            logger.debug("Renaming 'batch_size' hyperparam mutation to 'train_batch_size_per_learner'")
        self.hyperparam_mutations = mutations

    def configure(self) -> None:
        super().configure()
        self.add_argument(
            "--hyperparam-mutations",
            type=_to_hyperparam_mutations,
            default=None,
            help="Hyperparameter mutations for PopulationBasedTraining.",
        )
        # As long as Sentinel cannot be pickled do not add it as default
        # self.add_argument(
        #    "--metric",
        #    type=_resolve_default_metric,
        #    default=DEFAULT_EVAL_METRIC,
        #    help="The metric to be optimized, by default the evaluation metric.",
        # )

    def __init__(
        self,
        *args,
        underscores_to_dashes=False,
        explicit_bool=False,
        config_files=None,
        **kwargs,
    ) -> None:
        # TODO: config files if not available on remote
        super().__init__(
            *args,
            underscores_to_dashes=underscores_to_dashes,
            explicit_bool=explicit_bool,
            config_files=config_files,
            **kwargs,
        )

        # HACK: to_tap_class does not support overrides of default values
        # see https://github.com/swansonk14/typed-argument-parser/issues/166
        def replace_action(arg_name: str, new_default: Any) -> None:
            for action in self._actions:
                if action.dest == arg_name:
                    action.default = new_default
                    break

        for var, val in vars(PopulationBasedTrainingParser).items():
            # On python < 3.11 Sentinel passes callable() check
            if val is DEFAULT_EVAL_METRIC or not (
                var.startswith("_") or callable(val) or isinstance(val, (staticmethod, classmethod, property))
            ):
                replace_action(var, val)
        assert (action := next(a for a in self._actions if a.dest == "time_attr")).default == "current_step", (
            f"got {action.default}"
        )

    def to_scheduler(self) -> PopulationBasedTraining:
        if not self._parsed:
            args = self.parse_args().as_dict()
        else:
            args = self.as_dict()
        args.pop("hyperparam_mutations", None)  # will be set below
        # Set by Tuner, unset
        args["mode"] = None
        args["metric"] = None
        # non-scheduler args
        use_native = args.pop("use_native_pbt")
        if self.resample_probability >= 1.0 and self.hyperparam_mutations is None:
            raise ValueError("hyperparam_mutations must be set if resample_probability is 1.0")
        assert not TYPE_CHECKING or self.hyperparam_mutations is not None  # ray has implicit optional
        args = {arg: val for arg, val in args.items() if arg in inspect.signature(PopulationBasedTraining).parameters}

        if use_native:
            return PopulationBasedTraining(
                **args,
                hyperparam_mutations=self.hyperparam_mutations,
            )
        return TopPBTTrialScheduler(
            **args,
            hyperparam_mutations=self.hyperparam_mutations,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ray.tune.logger import LoggerCallback

from ray_utilities.postprocessing import save_videos

# ruff: noqa: ARG002
# pyright: reportArgumentType=false


if TYPE_CHECKING:
    from ray.tune.experiment.trial import Trial


class SaveVideoFirstCallback(LoggerCallback):
    def on_trial_result(
        self,
        iteration: int,
        trials: list["Trial"],
        trial: "Trial",
        result: dict[str, Any],
        **info,
    ):
        """Called after receiving a result from a trial.

        The search algorithm and scheduler are notified before this
        hook is called.

        Arguments:
            iteration: Number of iterations of the tuning loop.
            trials: List of trials.
            trial: Trial that just sent a result.
            result: Result that the trial sent.
            **info: Kwargs dict for forward compatibility.

        Saves videos to disk and provides a video_path to them
        """
        save_videos(result)
        super().on_trial_result(iteration, trials, trial, result, **info)



"""Video processing utilities for Ray RLlib experiment recordings.

This module provides utilities for converting NumPy arrays to video files, which is
essential for logging episode recordings from Ray RLlib experiments. The videos can
be used for visualizing agent behavior during training and evaluation.

The main functionality includes converting multi-dimensional NumPy arrays (typically
episode recordings) into standard video formats like MP4, with support for different
color channels and frame configurations.

Key Features:
    - Convert NumPy arrays to MP4 video files
    - Handle different array shapes and color channels
    - Temporary video file creation for logging
    - Compatible with various logging frameworks (Wandb, Comet, TensorBoard)

Example:
    Basic video creation from episode recording::

        import numpy as np
        from ray_utilities.video.numpy_to_video import create_temp_video

        # Episode frames: (time, height, width, channels)
        episode_frames = np.random.randint(0, 255, (100, 64, 64, 3), dtype=np.uint8)

        # Create temporary video file
        video_path = create_temp_video(episode_frames)
        print(f"Video saved to: {video_path}")

    Using with Ray RLlib callbacks::

        from ray_utilities.postprocessing import save_videos

        # In a callback, convert video arrays to files
        save_videos(training_results)  # Modifies results in-place

Functions:
    :func:`numpy_to_video`: Convert NumPy array to video file
    :func:`create_temp_video`: Create temporary video with unique filename

See Also:
    :mod:`ray_utilities.postprocessing`: For integrating videos into result processing
    :mod:`ray_utilities.constants`: Video-related constants and configurations
    :mod:`cv2`: OpenCV library used for video encoding
"""

from __future__ import annotations

import os
import uuid
from typing import TYPE_CHECKING, Optional

import cv2
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

from ray_utilities.temp_dir import TEMP_DIR_PATH


def numpy_to_video(
    data: list[float | int] | NDArray | list[NDArray], video_filename: str = "output.mp4", frame_rate: int = 10
) -> None:
    """Convert NumPy array data to an MP4 video file.

    This function takes episode recording data (typically from Ray RLlib environments)
    and converts it to a standard MP4 video file. It handles various input shapes
    and automatically transposes dimensions to match OpenCV's expected format.

    Args:
        data: Video data as NumPy array or list. Expected shapes:

            - **5D**: ``(batch=1, time, channel, height, width)`` - Standard Ray format
            - **4D**: ``(time, channel, height, width)`` or ``(time, height, width, channel)``
            - **List**: List of frame arrays

        video_filename: Output filename for the video. Should end with ``.mp4``.
        frame_rate: Frames per second for the output video.

    Note:
        **Shape Requirements**:

        Ray Tune's logging frameworks (WandB, TensorBoard) expect specific array shapes:

        - **3D array**: Single image ``(channel, height, width)``
        - **4D array**: Batch of images ``(batch, channel, height, width)``
        - **5D array**: Video ``(batch=1, time, channel, height, width)``

        This function automatically handles the conversion to the correct OpenCV format
        ``(time, height, width, channel)`` regardless of input shape.

    Example:
        >>> import numpy as np
        >>> # Create sample episode data (10 frames, 3 channels, 64x64)
        >>> frames = np.random.randint(0, 255, (10, 3, 64, 64), dtype=np.uint8)
        >>> numpy_to_video(frames, "episode.mp4", frame_rate=30)

        With 5D Ray format::

        >>> # Ray format: (1, time, channel, height, width)
        >>> ray_frames = np.random.randint(0, 255, (1, 20, 3, 128, 128), dtype=np.uint8)
        >>> numpy_to_video(ray_frames, "ray_episode.mp4")

    Raises:
        AssertionError: If the array doesn't have the expected channel dimensions
            (1 or 3 channels) in the correct positions.

    See Also:
        :func:`create_temp_video`: Wrapper that creates temporary filenames
        :mod:`cv2`: OpenCV video writing functionality
    """
    # Create a dummy video file in MP4 format
    video = np.squeeze(data)
    if video.shape[-1] not in (1, 3):
        assert video.shape[-3] in (1, 3)  # (L, C, H, W)
        # For CV2, the channel should be the last dimension
        video = video.transpose(0, 2, 3, 1)  # (L, H, W, C)
    _length, frame_height, frame_width, _colors = video.shape

    fourcc = cv2.VideoWriter_fourcc(*"avc1")  # type: ignore[attr-defined]
    out = cv2.VideoWriter(
        video_filename,
        fourcc=fourcc,
        fps=frame_rate,
        frameSize=(frame_width, frame_height),
    )

    if video.ndim == 4:
        for frame in video:
            out.write(frame)
    else:
        for frame in video.reshape(-1, *video.shape[-3:]):  # make 4D; might merges multiple videos
            out.write(frame)

    out.release()


def create_temp_video(
    data: list[float | int] | NDArray | list[NDArray],
    suffix: str = ".mp4",
    frame_rate: int = 10,
    dir: Optional[str] = None,
) -> str:
    """Create a temporary video file from NumPy array data with a unique filename.

    This is a convenience wrapper around :func:`numpy_to_video` that automatically
    generates a unique filename and saves the video to a temporary directory. It's
    particularly useful for creating videos that will be logged to experiment
    tracking platforms.

    Args:
        data: Video data as NumPy array or list. See :func:`numpy_to_video` for
            detailed shape requirements.
        suffix: File extension for the video. Currently only ``.mp4`` is supported.
        frame_rate: Frames per second for the output video.
        dir: Directory to save the video file. If ``None``, saves to the configured
            temporary directory (:data:`ray_utilities.temp_dir.TEMP_DIR_PATH`).

    Returns:
        Full path to the created temporary video file.

    Example:
        Create temporary video for logging::

            >>> import numpy as np
            >>> frames = np.random.randint(0, 255, (50, 64, 64, 3), dtype=np.uint8)
            >>> video_path = create_temp_video(frames, frame_rate=30)
            >>> print(video_path)
            '/tmp/ray_utilities/a1b2c3d4e5f6.mp4'

        Custom directory and frame rate::

            >>> video_path = create_temp_video(frames, dir="/my/videos", frame_rate=60)

    Note:
        - The filename is generated using :func:`uuid.uuid4` for uniqueness
        - Only MP4 format is currently supported
        - The temporary file is not automatically cleaned up

    See Also:
        :func:`numpy_to_video`: Underlying video conversion function
        :data:`ray_utilities.temp_dir.TEMP_DIR_PATH`: Default temporary directory
        :func:`ray_utilities.postprocessing.save_videos`: For batch video processing
    """
    assert suffix == ".mp4", "Only MP4 format is supported"
    path = os.path.join(TEMP_DIR_PATH if dir is None else dir, f"{uuid.uuid4().hex}{suffix}")
    numpy_to_video(data, path, frame_rate)
    return path



from __future__ import annotations

import json
import logging
import re
from typing import TYPE_CHECKING, TypeAlias

from ray.rllib.core.rl_module.multi_rl_module import MultiRLModule
from ray.rllib.core.rl_module.torch.torch_rl_module import TorchRLModule

if TYPE_CHECKING:
    from ray.rllib.algorithms.algorithm import Algorithm
    from ray.rllib.core.rl_module.rl_module import RLModule

logger = logging.getLogger(__name__)

_Node: TypeAlias = dict[str, str | dict[str, "_Node"] | list["_Node"]]


def _jsonify_summary(summary: str) -> dict:
    lines = [line.rstrip() for line in summary.strip().splitlines() if line.strip()]
    stack = []
    root: _Node = {}
    current: _Node = root
    indent_stack = [-1]
    name_stack = []

    def add_child(parent, name, child):
        if "children" not in parent:
            parent["children"] = {}
        parent["children"][name] = child

    for line in lines:
        indent = len(line) - len(line.lstrip())
        content = line.strip()

        # Match module start: e.g. (encoder): TorchActorCriticEncoder(
        m = re.match(r"\(([^)]+)\): ([^(]+)\($", content)
        if m:
            key, module = m.groups()
            node = {"type": module, "children": {}}
            add_child(current, key, node)
            stack.append(current)
            current = node
            indent_stack.append(indent)
            name_stack.append(key)
            continue

        # Match top-level module: e.g. DefaultPPOTorchRLModule(
        m = re.match(r"([^(]+)\($", content)
        if m:
            module = m.group(1)
            node = {"type": module}  # "children": {}
            root = node
            current = node
            stack = []
            indent_stack = [-1]
            name_stack = [module]
            continue

        # Match layer: e.g. (0): Linear(in_features=4, out_features=1, bias=True)
        m = re.match(r"\(([^)]+)\): ([^(]+)\((.*)\)", content)
        if m:
            key, layer, params = m.groups()
            if "layers" not in current:
                current["layers"] = []
            layer_info = {"name": key, "type": layer}
            if params:
                layer_info["params"] = params
            current["layers"].append(layer_info)  # pyright: ignore[reportAttributeAccessIssue]
            continue

        # Match simple layer: e.g. (1): Tanh()
        m = re.match(r"\(([^)]+)\): ([^(]+)\(\)", content)
        if m:
            key, layer = m.groups()
            if "layers" not in current:
                current["layers"] = []
            current["layers"].append({"name": key, "type": layer})  # pyright: ignore[reportAttributeAccessIssue]
            continue

        # Match block end: )
        if content == ")":
            while indent_stack and indent_stack[-1] >= indent:
                current = stack.pop()
                indent_stack.pop()
                name_stack.pop()
            continue

    return root


def save_model_config_and_architecture(*, algorithm: "Algorithm", **kwargs) -> None:
    """on_algorithm_init callback that saves the model config and architecture as json dict."""
    module = _get_module(algorithm)
    config = _get_module_config(module)
    for k, v in config.items():
        # Step 1: Convert value to string and replace escaped newlines
        value_str = repr(v).replace("\\n", "\n")
        # Step 2: Replace multiple spaces with ', '
        value_with_commas = re.sub(r"\ {2,}", ", ", value_str)
        # Step 3: Replace 'inf ' with 'inf,'
        cleaned_value = re.sub(r"inf ", "inf,", value_with_commas)
        config[k] = cleaned_value

    arch = _get_model_architecture(module)
    output = {
        "config": config,
        "architecture": arch,
    }
    out_path = "./model_architecture.json"
    try:
        with open(out_path, "w") as f:
            json.dump(output, f, indent=2)
        logger.info("Saved model architecture/config to: %s", out_path)
    except OSError as e:
        logger.error("Failed to save model architecture/config: %s", str(e))


def _get_module(algorithm: "Algorithm") -> TorchRLModule:
    module = getattr(algorithm, "rl_module", None)
    if module is None:
        try:
            module = algorithm.learner_group._learner.module  # pyright: ignore[reportOptionalMemberAccess]
            assert module
        except AttributeError:
            module = algorithm.config.rl_module_spec.build()  # pyright: ignore[reportOptionalMemberAccess]
    if isinstance(module, MultiRLModule):
        module = module["default_policy"]
    if isinstance(module, TorchRLModule):
        return module
    if module is not None and hasattr(module, "_modules"):
        modules = getattr(module, "_modules", {})
        for m in modules.values():
            if isinstance(m, TorchRLModule):
                return m
    raise RuntimeError("No TorchRLModule found in algorithm.rl_module")


def _get_module_config(module: TorchRLModule | RLModule) -> dict:
    # config of RLModule is deprecated
    constructor_params = module.get_ctor_args_and_kwargs()
    return {"args": constructor_params[0], **constructor_params[1]}


def _get_model_architecture(module: TorchRLModule) -> dict:
    arch = {}
    torch_model = getattr(module, "model", module)
    arch["summary_str"] = str(torch_model)
    arch["summary_json"] = _jsonify_summary(arch["summary_str"])
    arch["layers"] = _extract_layers(torch_model)
    return arch


def _extract_layers(torch_model) -> list:
    layers = []
    for name, layer in getattr(torch_model, "named_modules", list)():
        if name == "":
            continue
        layer_info = {
            "name": name,
            "type": layer.__class__.__name__,
            "params": sum(p.numel() for p in getattr(layer, "parameters", lambda recurse=False: [])(recurse=False)),
        }
        layers.append(layer_info)
    return layers



"""Constants and configuration values used throughout Ray Utilities.

Defines important constants, version checks, and configuration values used across
the Ray Utilities package. Includes version compatibility flags, metric keys for
Ray RLlib, Comet ML configuration, and video logging constants.

Key Constants:
    :data:`RAY_VERSION`: Current Ray version for compatibility checks
    :data:`RAY_NEW_API_STACK_ENABLED`: Whether Ray's new API stack is available
    :data:`EVAL_METRIC_RETURN_MEAN`: Standard evaluation return metric key
    :data:`DEFAULT_VIDEO_DICT_KEYS`: Video logging configuration keys
    :data:`COMET_OFFLINE_DIRECTORY`: Path for offline Comet ML experiments

Example:
    >>> from ray_utilities.constants import RAY_NEW_API_STACK_ENABLED
    >>> if RAY_NEW_API_STACK_ENABLED:
    ...     # Use new Ray API features
    ...     pass
"""
# pyright: enableExperimentalFeatures=true

import logging
import re
from typing import Final

import gymnasium as gym
import ray
from packaging.version import Version
from packaging.version import parse as parse_version
from ray.rllib.core import DEFAULT_AGENT_ID, DEFAULT_MODULE_ID
from ray.rllib.utils.metrics import (
    ENV_RUNNER_RESULTS,
    EPISODE_DURATION_SEC_MEAN,
    EPISODE_LEN_MAX,
    EPISODE_LEN_MEAN,
    EPISODE_LEN_MIN,
    EPISODE_RETURN_MAX,
    EPISODE_RETURN_MEAN,
    EPISODE_RETURN_MIN,
    EVALUATION_RESULTS,
)
from typing_extensions import Sentinel

from ._runtime_constants import (
    COMET_OFFLINE_DIRECTORY,
    ENTRY_POINT,
    ENTRY_POINT_ID,
    RAY_UTILITIES_INITIALIZATION_TIMESTAMP,
    RUN_ID,
)

__all__ = []
__all__ += [
    "COMET_OFFLINE_DIRECTORY",
    "ENTRY_POINT",
    "ENTRY_POINT_ID",
    "RAY_UTILITIES_INITIALIZATION_TIMESTAMP",
    "RUN_ID",
]

_logger = logging.getLogger(__name__)

# region runtime constants

# Version Compatibility Flags
RAY_VERSION = parse_version(ray.__version__)
"""Version: Parsed version of the currently installed Ray package.

Used throughout the codebase for version-specific compatibility checks and feature detection.

Example:
    >>> from ray_utilities.constants import RAY_VERSION
    >>> if RAY_VERSION >= Version("2.40.0"):
    ...     print("New API stack available")
"""

GYM_VERSION = parse_version(gym.__version__)
"""Version: Parsed version of the currently installed Gymnasium package."""

GYM_V1: bool = GYM_VERSION >= Version("1.0.0")
"""bool: True if Gymnasium version 1.0.0 or higher is installed.

Gymnasium 1.0.0 introduced significant API changes and improvements over earlier versions.
"""

GYM_V_0_26: bool = GYM_VERSION >= Version("0.26")
"""bool: True if Gymnasium version 0.26 or higher is installed.

Version 0.26 was the first official Gymnasium release after the transition from OpenAI Gym.
This flag is used to handle compatibility between legacy Gym and modern Gymnasium.
"""

RAY_NEW_API_STACK_ENABLED = RAY_VERSION >= Version("2.40.0")
"""bool: True if Ray's new API stack is available (Ray >= 2.40.0).

The new API stack introduced significant improvements to RLlib's architecture,
including better modularity and performance. This flag enables conditional
code paths for new vs. legacy API usage.

See Also:
    `Ray RLlib New API Stack Migration Guide
    <https://docs.ray.io/en/latest/rllib/new-api-stack-migration-guide.html>`_
"""

# endregion

RAY_UTILITIES_NEW_LOG_FORMAT: Final[str] = "RAY_UTILITIES_NEW_LOG_FORMAT"
"""
Environment variable key to enable new log format if found in :py:obj:`os.environ`.

Note:
    The new logging format is used in downstream loggers, e.g. for WandB or Comet ML.
    other callbacks and logs are not affected and should not check for the new format.
"""

TRAINING = "training"
"""For the rllib layout use ENV_RUNNER_RESULTS instead of "training"."""

# Evaluation and Training Metric Keys
EVAL_METRIC_RETURN_MEAN = EVALUATION_RESULTS + "/" + ENV_RUNNER_RESULTS + "/" + EPISODE_RETURN_MEAN
"""str: Standard path for evaluation episode return mean metric.

This metric key path follows Ray RLlib's hierarchical result structure:
``evaluation/env_runner_results/episode_return_mean``
"""

NEW_LOG_EVAL_METRIC = EVALUATION_RESULTS + "/" + EPISODE_RETURN_MEAN
"""Metric key used in loggers when new log format is used."""

DISC_EVAL_METRIC_RETURN_MEAN = EVALUATION_RESULTS + "/discrete/" + ENV_RUNNER_RESULTS + "/" + EPISODE_RETURN_MEAN
"""str: Path for discrete evaluation episode return mean metric.

Used when discrete evaluation is enabled alongside standard evaluation:
``evaluation/discrete/env_runner_results/episode_return_mean``
"""

NEW_LOG_DISC_EVAL_METRIC = EVALUATION_RESULTS + "/discrete/" + EPISODE_RETURN_MEAN
"""Metric key used in loggers when new log format is used and discrete evaluation is enabled."""

TRAIN_METRIC_RETURN_MEAN = ENV_RUNNER_RESULTS + "/" + EPISODE_RETURN_MEAN
"""str: Standard path for training episode return mean metric.

Training metric path: ``env_runner_results/episode_return_mean``
"""

NEW_LOG_TRAIN_METRIC = TRAINING + "/" + EPISODE_RETURN_MEAN
"""Metric key used in loggers when new log format is used and training is enabled."""

DEFAULT_EVAL_METRIC = Sentinel("DEFAULT_EVAL_METRIC")
"""typing.Sentinel value

Default evaluation metric sentinel value to be replaced with actual metric key
depending on wether the new log format is used or not (variable RAY_UTILITIES_NEW_LOG_FORMAT is set),
also possibly depending on wether discrete evaluation is enabled or not.

The correct metric key should be retrieved by:

.. code-block:: python

    if metric is DEFAULT_EVAL_METRIC:
        metric = NEW_LOG_EVAL_METRIC if new_log_format_used() else EVAL_METRIC_RETURN_MEAN
"""

ENVIRONMENT_RESULTS = "environments"
"""
Metric key for environment-specific results in Trainable results.

See Also:
    :class:`SeededEnvsCallback`
"""

# Video Recording Constants
EPISODE_VIDEO_PREFIX = "episode_videos_"
"""str: Prefix used for all episode video metric keys."""

EPISODE_BEST_VIDEO = "episode_videos_best"
"""str: Key for best episode video recordings."""

EPISODE_WORST_VIDEO = "episode_videos_worst"
"""str: Key for worst episode video recordings."""

EVALUATION_BEST_VIDEO_KEYS = (EVALUATION_RESULTS, ENV_RUNNER_RESULTS, EPISODE_BEST_VIDEO)
EVALUATION_WORST_VIDEO_KEYS = (EVALUATION_RESULTS, ENV_RUNNER_RESULTS, EPISODE_WORST_VIDEO)
DISCRETE_EVALUATION_BEST_VIDEO_KEYS = (EVALUATION_RESULTS, "discrete", ENV_RUNNER_RESULTS, EPISODE_BEST_VIDEO)
DISCRETE_EVALUATION_WORST_VIDEO_KEYS = (EVALUATION_RESULTS, "discrete", ENV_RUNNER_RESULTS, EPISODE_WORST_VIDEO)

EVALUATION_BEST_VIDEO = "/".join(EVALUATION_BEST_VIDEO_KEYS)
EVALUATION_WORST_VIDEO = "/".join(EVALUATION_WORST_VIDEO_KEYS)
DISCRETE_EVALUATION_BEST_VIDEO = "/".join(DISCRETE_EVALUATION_BEST_VIDEO_KEYS)
DISCRETE_EVALUATION_WORST_VIDEO = "/".join(DISCRETE_EVALUATION_WORST_VIDEO_KEYS)

# region: new layout for log metrics

# NEW variants: omit ENV_RUNNER_RESULTS for evaluation, substitute ENV_RUNNER_RESULTS with "training" for training
EVALUATION_BEST_VIDEO_KEYS_NEW = (EVALUATION_RESULTS, EPISODE_BEST_VIDEO)
EVALUATION_WORST_VIDEO_KEYS_NEW = (EVALUATION_RESULTS, EPISODE_WORST_VIDEO)
DISCRETE_EVALUATION_BEST_VIDEO_KEYS_NEW = (EVALUATION_RESULTS, "discrete", EPISODE_BEST_VIDEO)
DISCRETE_EVALUATION_WORST_VIDEO_KEYS_NEW = (EVALUATION_RESULTS, "discrete", EPISODE_WORST_VIDEO)

EVALUATION_BEST_VIDEO_NEW = "/".join(EVALUATION_BEST_VIDEO_KEYS_NEW)
EVALUATION_WORST_VIDEO_NEW = "/".join(EVALUATION_WORST_VIDEO_KEYS_NEW)
DISCRETE_EVALUATION_BEST_VIDEO_NEW = "/".join(DISCRETE_EVALUATION_BEST_VIDEO_KEYS_NEW)
DISCRETE_EVALUATION_WORST_VIDEO_NEW = "/".join(DISCRETE_EVALUATION_WORST_VIDEO_KEYS_NEW)

# NEW variants for video dict keys (evaluation omits ENV_RUNNER_RESULTS, training uses "training")
DEFAULT_VIDEO_DICT_KEYS_NEW = (
    EVALUATION_BEST_VIDEO_KEYS_NEW,
    EVALUATION_WORST_VIDEO_KEYS_NEW,
    DISCRETE_EVALUATION_BEST_VIDEO_KEYS_NEW,
    DISCRETE_EVALUATION_WORST_VIDEO_KEYS_NEW,
)
"""tuple[tuple[str, ...], ...]: Collection of new-style video key tuples for logging.

Evaluation keys omit ENV_RUNNER_RESULTS, training keys use "training" instead.
"""

DEFAULT_VIDEO_DICT_KEYS_FLATTENED_NEW = (
    EVALUATION_BEST_VIDEO_NEW,
    EVALUATION_WORST_VIDEO_NEW,
    DISCRETE_EVALUATION_BEST_VIDEO_NEW,
    DISCRETE_EVALUATION_WORST_VIDEO_NEW,
)
"""tuple[str, ...]: Flattened string keys for new-style video logging.

Evaluation keys omit ENV_RUNNER_RESULTS, training keys use "training" instead.
"""

# endregion

DEFAULT_VIDEO_DICT_KEYS = (
    *DEFAULT_VIDEO_DICT_KEYS_NEW,
    EVALUATION_BEST_VIDEO_KEYS,
    EVALUATION_WORST_VIDEO_KEYS,
    DISCRETE_EVALUATION_BEST_VIDEO_KEYS,
    DISCRETE_EVALUATION_WORST_VIDEO_KEYS,
)
"""tuple[tuple[str, ...], ...]: Collection of video key tuples for default video logging.

This contains the hierarchical key paths for all default video types that can be logged
during evaluation. Each tuple represents a path through the nested result dictionary.

Note:
    The video data might be a dictionary containing both ``"video"`` and ``"reward"`` keys
    rather than just the raw video array.

See Also:
    :data:`DEFAULT_VIDEO_DICT_KEYS_FLATTENED`: String versions of these keys
"""

DEFAULT_VIDEO_DICT_KEYS_FLATTENED = (
    *DEFAULT_VIDEO_DICT_KEYS_FLATTENED_NEW,
    EVALUATION_BEST_VIDEO,
    EVALUATION_WORST_VIDEO,
    DISCRETE_EVALUATION_BEST_VIDEO,
    DISCRETE_EVALUATION_WORST_VIDEO,
)
"""tuple[str, ...]: Flattened string keys for default video logging.

These are the slash-separated string versions of the video keys for use in flat
dictionaries or when the nested structure has been flattened.

Note:
    The video data might be a dictionary containing both ``"video"`` and ``"reward"`` keys
    rather than just the raw video array.

See Also:
    :data:`DEFAULT_VIDEO_DICT_KEYS`: Tuple versions of these keys for nested access
"""

assert all(EPISODE_VIDEO_PREFIX in key for key in DEFAULT_VIDEO_DICT_KEYS_FLATTENED)

EVALUATED_THIS_STEP = "evaluated_this_step"
"""str: Boolean metric key to indicate evaluation was performed this training step.

When logged with ``reduce_on_results=True``, this metric tracks whether evaluation
was actually run during a particular training iteration, which is useful for
conditional processing of evaluation results.
"""


# CLI and Reporting Configuration
CLI_REPORTER_PARAMETER_COLUMNS = ["algo", "module", "model_config"]
"""list[str]: Default parameter columns to display in CLI progress reports.

These parameter keys from the search space will be shown in Ray Tune's command-line
progress reporter for easier experiment monitoring.
"""

# Sampling and Training Metrics
NUM_ENV_STEPS_PASSED_TO_LEARNER = "num_env_steps_passed_to_learner"
"""str: Metric key for environment steps passed to learner with exact sampling.

When using exact sampling mode, this tracks the actual number of environment steps
that were passed to the learner during the current training iteration, which may
differ from the total steps sampled due to exact sampling constraints.
"""

NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME = "num_env_steps_passed_to_learner_lifetime"
"""str: Lifetime metric for environment steps passed to learner with exact sampling.

Cumulative count of environment steps passed to the learner over the entire
training run when using exact sampling mode.
"""

CURRENT_STEP = "current_step"
"""str: The current training step metric key.

This top-level metric tracks the current step in the training process. With exact
sampling, it aligns with :data:`NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME`, otherwise
it aligns with ``NUM_ENV_STEPS_SAMPLED_LIFETIME``.

Used for consistent step tracking across different sampling modes and is the
recommended metric for tracking training progress.
"""

PERTURBED_HPARAMS = "__perturbed__"
"""str: Configuration key indicating which hyperparameters have been perturbed.

This key in a trainable's config dictionary contains information about which
hyperparameters have been modified from their original values. It's primarily
used by :meth:`~ray_utilities.training.default_class.DefaultTrainable.load_checkpoint`
and not during trainable initialization.

Note:
    This is used internally by the checkpoint/restore system and typically should
    not be set manually in experiment configurations.
"""

EPISODE_METRICS_KEYS = (
    EPISODE_DURATION_SEC_MEAN,
    EPISODE_LEN_MAX,
    EPISODE_LEN_MEAN,
    EPISODE_LEN_MIN,
    EPISODE_RETURN_MAX,
    EPISODE_RETURN_MEAN,
    EPISODE_RETURN_MIN,
    ("module_episode_return_mean", DEFAULT_MODULE_ID),
    ("agent_episode_return_mean", DEFAULT_AGENT_ID),
)
"""Keys that are by default logged with a window by RLlib. When using"""

FORK_FROM = "fork_from"
"""
Key in trial configs to indicate the trial id and optionally step a trial was forked from.

The value should have the format ``<trial_id>?_step=<step>`` where ``?_step=<step>`` is optional,
it can be parsed with :const:`RE_PARSE_FORK_FROM` (deprecated)
or be a dict defined as :class:`ForkFromData`.

Note:
    A trial should never contain the ``?`` character.
"""

RE_PARSE_FORK_FROM = re.compile(r"^(?P<fork_id>[^?]*?)(?:\?_step=(?P<fork_step>\d+))?$")
"""
Regex pattern to parse the value of :const:`FORK_FROM` into trial id and step.

The ``fork_id`` will be everything before the first ``?``, the ``?_step=<fork_step>`` part is optional.
"""



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Protocol

from ray.rllib.core.models.torch.base import TorchModel  # noqa: F401  # type: ignore
from typing_extensions import Self

if TYPE_CHECKING:
    from ray.rllib.core.rl_module.rl_module import RLModuleConfig


class DiscreteModelABC(ABC):
    """
    Model that implements `create_discrete_copy` for discrete evaluation

    As an alternative use :class:`DiscreteModuleBase`, that implements :meth:`~DiscreteModuleBase.switch_mode`
    but requires a different logging approach.
    """

    @abstractmethod
    def create_discrete_copy(self) -> Self: ...


class _DiscreteTorchModelBase(DiscreteModelABC, TorchModel): ...


if TYPE_CHECKING:
    import os
    from ray.rllib.core.models.tf.base import TfModel  # pyright: ignore[reportMissingImports] # ray has removed this somewhere around 2.40

    if not os.environ.get("SPHINX_BUILD", "0") == "1":
        from typing import type_check_only
    else:
        type_check_only = lambda x: x  # noqa: E731

    @type_check_only
    class _DiscreteTFModelBase(DiscreteModelABC, TfModel): ...


class DiscreteModuleBase(Protocol):
    CAN_USE_DISCRETE_EVAL = True

    if TYPE_CHECKING:

        def __instance_members(self):
            self.is_discrete: bool
            self.inference_only: bool
            self.config: RLModuleConfig

    @abstractmethod
    def switch_mode(self, *, discrete: bool): ...


class _TorchPPOModule(DiscreteModuleBase):
    def __instance_members(self):
        super().__instance_members()
        self.pi: _DiscreteTorchModelBase
        self.vf: _DiscreteTorchModelBase


class _TFPPOModule(DiscreteModuleBase):
    def __instance_members(self):
        super().__instance_members()
        self.pi: _DiscreteTFModelBase
        self.vf: _DiscreteTFModelBase


class DiscreteTorchPPOModule(_TorchPPOModule, DiscreteModuleBase):
    def switch_mode(self, *, discrete: bool):
        assert self.inference_only == self.config.inference_only
        if discrete and not self.is_discrete:
            self.pi = self.pi.create_discrete_copy()
            self.pi.eval()
            if not self.inference_only:  # vf is not used in inference and missing in later ray versions
                self.vf = self.vf.create_discrete_copy()
                self.vf.eval()
            self.is_discrete = True
        elif not discrete and self.is_discrete:
            self.pi = self.pi
            if not self.inference_only:
                self.vf = self.vf
            self.is_discrete = False


class DiscreteTFPPOModule(_TFPPOModule, DiscreteModuleBase):
    def switch_mode(self, *, discrete: bool):
        assert self.inference_only == self.config.inference_only
        if discrete and not self.is_discrete:
            self.pi = self.pi.create_discrete_copy()
            # self.pi.eval()
            if not self.inference_only:  # vf is not used in inference and missing in later ray versions
                self.vf = self.vf.create_discrete_copy()
                # self.vf.eval()
            self.is_discrete = True
        elif not discrete and self.is_discrete:
            self.pi = self.pi
            if not self.inference_only:
                self.vf = self.vf
            self.is_discrete = False



"""Type definitions and aliases for Ray Utilities components.

Provides type definitions for Ray RLlib return data, metrics dictionaries,
trainable configurations, and other commonly used types throughout the package.

Main Type Aliases:
    - :data:`AlgorithmReturnData`: Ray RLlib algorithm training results
    - :data:`TrainableReturnData`: Ray Tune trainable return structure
    - :data:`LogMetricsDict`: Structured metrics for logging frameworks
    - :data:`RewardsDict`: Episode reward tracking structure

Example:
    >>> from ray_utilities.typing import AlgorithmReturnData
    >>> def process_results(results: AlgorithmReturnData) -> float:
    ...     return results["env_runner_results"]["episode_return_mean"]

See Also:
    :mod:`ray_utilities.typing.algorithm_return`: Algorithm-specific return types
    :mod:`ray_utilities.typing.trainable_return`: Trainable return structures
    :mod:`ray_utilities.typing.metrics`: Metrics and logging type definitions
    :mod:`typing_extensions`: Extended typing capabilities
"""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Iterable, Literal, Mapping, NamedTuple

import typing_extensions
from typing_extensions import NotRequired, TypeAliasType, TypedDict, TypeVar

from .algorithm_return import (
    AlgorithmReturnData,
    EnvRunnersResultsDict,
    EvaluationResultsDict,
    StrictAlgorithmReturnData,
)
from .common import (
    BaseEnvRunnersResultsDict,
    BaseEvaluationResultsDict,
    VideoTypes,
)
from .metrics import FlatLogMetricsDict, LogMetricsDict
from .trainable_return import RewardsDict, RewardUpdaters, TrainableReturnData

if TYPE_CHECKING:
    import gymnasium as gym
    import ray.tune
    from gymnasium.envs.registration import EnvSpec as _EnvSpec
    from gymnax.environments import environment as environment_gymnax
    from ray.rllib.algorithms import Algorithm, AlgorithmConfig
    from ray.tune.experiment import Trial
    from ray.tune.search.sample import Domain

    from ray_utilities.setup.experiment_base import ExperimentSetupBase


__all__ = [
    "AlgorithmReturnData",
    "BaseEnvRunnersResultsDict",
    "BaseEvaluationResultsDict",
    "CometStripedVideoFilename",
    "EnvRunnersResultsDict",
    "EvaluationResultsDict",
    "FlatLogMetricsDict",
    "FunctionalTrainable",
    "LogMetricsDict",
    "RewardUpdaters",
    "RewardsDict",
    "StrictAlgorithmReturnData",
    "TestModeCallable",
    "TrainableReturnData",
    "VideoTypes",
]


CometStripedVideoFilename = Literal[
    "evaluation_best_video",
    "evaluation_discrete_best_video",
    "evaluation_worst_video",
    "evaluation_discrete_worst_video",
]

FunctionalTrainable = typing_extensions.TypeAliasType(
    "FunctionalTrainable", Callable[[dict[str, Any]], TrainableReturnData]
)

_ESB_co = TypeVar("_ESB_co", bound="ExperimentSetupBase[Any, AlgorithmConfig, Algorithm]", covariant=True)

TestModeCallable = typing_extensions.TypeAliasType(
    "TestModeCallable", Callable[[FunctionalTrainable, _ESB_co], TrainableReturnData], type_params=(_ESB_co,)
)

EnvSpec = TypeAliasType("EnvSpec", "str | _EnvSpec")
EnvType = TypeAliasType("EnvType", "gym.Env | environment_gymnax.Environment | gym.vector.VectorEnv")

_Domain = TypeVar("_Domain", bound="Domain", default="Domain")
_AnyT = TypeVar("_AnyT", default=Any)
ParameterSpace = TypeAliasType(
    "ParameterSpace",
    dict[Literal["grid_search"] | str, Iterable[_AnyT]] | _Domain,  # noqa: PYI051
    type_params=(_AnyT, _Domain),
)
"""Describes a tune.Domain or grid_search for a parameter sampling by tune"""

StopperType = TypeAliasType("StopperType", "Mapping | ray.tune.Stopper | Callable[[str, Mapping], bool] | None")


class Forktime(NamedTuple):
    time_attr: str
    """String to describe what the :attr:`time` tracks, e.g. ``training_iteration`` or ``current_step``"""
    time: int | float
    """Current time of the fork, tracked in :attr:`time_attr` units"""


class ForkFromData(TypedDict):
    parent_id: str
    """Trial id of the run to fork"""

    parent_training_iteration: int
    """Training iteration the fork is at. This is needed for example for WandB's fork_from feature"""

    parent_time: Forktime
    """
    Current time the fork is at.

    First element of the :class:`Forktime` is a string that describes the unit,
    second the numerical value of the time
    """

    parent_trial: NotRequired[Trial]
    """
    If available, the actual Trial object of the run to fork

    Attention:
        The trial might be modified during the tuning process
        as itself could be forked at some time.
    """

    parent_fork_id: NotRequired[str]
    """If available, the fork_id of the parent trial, if it was forked itself"""

    fork_id_this_trial: NotRequired[str]
    """The fork_id this trial will get. It is constructed from the other fields."""

    controller: NotRequired[str | Any]
    """A field to inform who decided the forking, e.g. scheduler name"""



"""Advanced Population Based Training scheduler for Ray Tune hyperparameter optimization.

This module provides :class:`ReTuneScheduler`, an enhanced version of Ray Tune's
PopulationBasedTraining that supports grid search mutations, flexible quantile
fractions, and improved trial management for reinforcement learning experiments.

Key Components:
    - :class:`ReTuneScheduler`: Enhanced PBT scheduler with advanced features
    - :func:`_grid_search_sample_function`: Grid search sampling utilities
    - Integration with Ray Tune's trial and search algorithm framework

The scheduler extends the standard PBT approach with support for deterministic
grid search mutations and more flexible population management strategies.
"""

from __future__ import annotations

import logging
import math
import random
from functools import partial
from itertools import cycle
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, TypeVar

from ray.train._internal.session import _FutureTrainingResult
from ray.tune.experiment import Trial
from ray.tune.schedulers.pbt import PopulationBasedTraining
from ray.util.debug import log_once

from ray_utilities.constants import PERTURBED_HPARAMS

if TYPE_CHECKING:
    from collections.abc import Iterable, MutableMapping

    from ray.tune.execution.tune_controller import TuneController
    from ray.tune.schedulers.pbt import _PBTTrialState
    from ray.tune.search.sample import Domain


logger = logging.getLogger(__name__)
_T = TypeVar("_T")


def _grid_search_sample_function(grid_search_space: Iterable[_T], *, repeat=True) -> Callable[[], _T]:
    """Create a function for sampling from a grid search space.

    Returns a parameterless function that yields grid search values either cyclically
    (with repetition) or once through the space (without repetition).

    Args:
        grid_search_space: Iterable containing the values to sample from.
        repeat: If True, cycle through values infinitely. If False, each value
            is returned once and then StopIteration is raised.

    Returns:
        A parameterless function that returns the next grid search sample.
        When repeat=False, the function raises StopIteration after all values
        have been returned once.

    Example:
        >>> sampler = _grid_search_sample_function([1, 2, 3], repeat=True)
        >>> sampler()  # Returns 1
        >>> sampler()  # Returns 2
        >>> sampler()  # Returns 3
        >>> sampler()  # Returns 1 (cycles back)
    """
    if repeat:
        cycler = cycle(grid_search_space)

        def cyclic_grid_iterator():
            return next(cycler)

        return cyclic_grid_iterator
    grid_search_space = list(grid_search_space)

    def grid_iterator():
        try:
            return grid_search_space.pop(0)
        except IndexError as e:
            raise StopIteration from e

    return grid_iterator


def _debug_dump_new_config(new_config: dict, mutation_keys: list[str]):
    logger.info("New config after perturbation %s", new_config)
    new_config[PERTURBED_HPARAMS] = {k: new_config[k] for k in mutation_keys}
    return new_config


class ReTuneScheduler(PopulationBasedTraining):
    """Enhanced Population Based Training scheduler with grid search and flexible quantiles.

    This scheduler extends Ray Tune's PopulationBasedTraining with support for grid search
    mutations, quantile fractions above 0.5, and improved trial management for reinforcement
    learning experiments.

    Key enhancements over standard PBT:
        - Grid search mutations for deterministic hyperparameter exploration
        - Support for quantile fractions > 0.5 for exploit-heavy strategies
        - Custom exploration functions with mutation tracking
        - Enhanced logging and debugging capabilities

    The scheduler maintains compatibility with the standard PBT interface while providing
    additional flexibility for advanced hyperparameter optimization strategies.

    Args:
        time_attr: Attribute to use for time progression tracking.
        metric: Metric name to optimize (e.g., "episode_reward_mean").
        mode: Optimization mode, either "max" or "min".
        perturbation_interval: Number of time units between perturbations.
        burn_in_period: Time units before perturbations begin.
        hyperparam_mutations: Dictionary mapping hyperparameter names to mutation
            specifications. Supports grid_search definitions for deterministic sampling.
        quantile_fraction: Fraction of population to consider as "top performers".
            Unlike standard PBT, values > 0.5 are supported for exploit-heavy strategies.
        resample_probability: Probability of resampling parameters during perturbation.
        perturbation_factors: Tuple of (lower, upper) factors for parameter perturbation.
        custom_explore_fn: Optional custom function for exploration logic.
        log_config: Whether to log configuration changes.
        require_attrs: Whether to require time_attr and metric in results.
        synch: Whether to use synchronous perturbation.

    Example:
        >>> scheduler = ReTuneScheduler(
        ...     metric="episode_reward_mean",
        ...     mode="max",
        ...     perturbation_interval=50000,
        ...     hyperparam_mutations={
        ...         "lr": {"grid_search": [1e-4, 5e-4, 1e-3]},
        ...         "batch_size": {"grid_search": [64, 128, 256]},
        ...     },
        ...     quantile_fraction=0.8,  # Exploit top 80% of trials
        ... )

    Note:
        Grid search spaces in hyperparam_mutations are automatically converted to
        sampling functions that cycle through the provided values.

    See Also:
        :class:`ray.tune.schedulers.pbt.PopulationBasedTraining`: Base PBT scheduler
        :func:`_grid_search_sample_function`: Grid search sampling utilities
    """

    def __init__(
        self,
        *,
        time_attr: str = "current_step",
        metric: str | None = None,
        mode: str | None = "max",
        perturbation_interval: float = 100_000,
        burn_in_period: float = 0,
        hyperparam_mutations: Optional[
            MutableMapping[str, dict[str, Any] | list | tuple | Callable[..., Any] | Domain]
        ] = None,
        # Use only very best trial # TODO: Should probably use more but double trials.
        quantile_fraction: float = 0.99,  # 0.25,  # 0 for no exploit -> no top trials, 0.99 for only exploit top trial
        resample_probability: float = 1.0,  # Always resample, assume grid_search in hyperparam_mutations # TODO: alt use custom_explore_fn with new value as input
        perturbation_factors: Tuple[float, float] = (0.8, 1.2),
        custom_explore_fn: Callable[..., Any] | None = None,
        log_config: bool = True,
        require_attrs: bool = True,
        synch: bool = False,
    ):
        if custom_explore_fn is None:  # Otherwise use a wrapper
            # XXX
            if hyperparam_mutations:
                custom_explore_fn = partial(_debug_dump_new_config, mutation_keys=list(hyperparam_mutations.keys()))
        # TODO: Do we want to reload the env_seed?
        self._trial_state: dict[Trial, _PBTTrialState]
        if quantile_fraction > 0.5:
            if quantile_fraction > 1:
                raise ValueError("quantile_fraction may not be larger than 1.0")
            quantile_fraction_after_init = quantile_fraction
            quantile_fraction = 0.5
        else:
            quantile_fraction_after_init = None
        if hyperparam_mutations:  # either hyperparam_mutations or custom_explore_fn must be passed
            for k, v in hyperparam_mutations.items():
                if isinstance(v, dict) and "grid_search" in v:
                    hyperparam_mutations[k] = _grid_search_sample_function(v["grid_search"])
        super().__init__(
            time_attr,
            metric,
            mode,
            perturbation_interval,
            burn_in_period,
            hyperparam_mutations,  # pyright: ignore[reportArgumentType]
            quantile_fraction,
            resample_probability,
            perturbation_factors,
            custom_explore_fn,  # only used on explore (see _exploit function, get_new_config)
            log_config,
            require_attrs,
            synch,
        )
        if quantile_fraction_after_init is not None:
            # ray code does not allow _quantile_fraction > 0.5
            self._quantile_fraction = quantile_fraction_after_init

    def on_trial_add(self, tune_controller: TuneController, trial: Trial):
        """Updates the trials config with hyperparam_mutations"""
        # Adds a new trial with config updated based on hyperparam_mutations
        return super().on_trial_add(tune_controller, trial)

    def on_trial_resultX(self, tune_controller: TuneController, trial: Trial, result: Dict) -> str:
        self._check_result(result)
        if self._metric not in result or self._time_attr not in result:
            return self.CONTINUE  # todo ray use Enum
        time = result[self._time_attr]
        state = self._trial_state[trial]
        # Continue training if burn-in period has not been reached, yet.
        if time < self._burn_in_period:
            logger.debug("Still in burn-in period: %s < %s", time, self._burn_in_period)
            return self.CONTINUE

        # Continue training if perturbation interval has not been reached, yet.
        time_since_perturb = time - state.last_perturbation_time
        if time_since_perturb < self._perturbation_interval:
            logger.debug("Perturbation interval not reached: %s < %s", time_since_perturb, self._perturbation_interval)
            return self.CONTINUE  # avoid checkpoint overhead

        logger.debug("Updating trial state for trial %s at time %s", trial, time)
        # update internal information, does not create a checkpoint!
        self._save_trial_state(state, time, result, trial)

        if not self._synch:
            state.last_perturbation_time = time
            # Divide executed trials in upper_quantile to safe
            lower_quantile, upper_quantile = self._quantiles()
            decision = self.CONTINUE
            for other_trial in tune_controller.get_trials():
                if other_trial.status in [Trial.PENDING, Trial.PAUSED]:
                    decision = self.PAUSE
                    break
            self._checkpoint_or_exploit(trial, tune_controller, upper_quantile, lower_quantile)
            return self.NOOP if trial.status == Trial.PAUSED else decision

        # calls self._checkpoint_or_exploit  # current_step (- last perturbation time) < self._perturbation_interval
        return super().on_trial_result(tune_controller, trial, result)

    def _quantiles(self) -> Tuple[List[Trial], List[Trial]]:
        """Returns trials in the lower and upper `quantile` of the population.

        If there is not enough data to compute this, returns empty lists.

        This method allows for quantile_fraction > 0.5 as well.
        """
        if self._quantile_fraction <= 0.5:
            return super()._quantiles()
        trials = []
        for trial, state in self._trial_state.items():
            logger.debug("Trial %s, state %s", trial, state)
            if trial.is_finished():
                logger.debug("Trial %s is finished", trial)
            if state.last_score is not None and not math.isnan(state.last_score) and not trial.is_finished():
                trials.append(trial)
        # called after trial_save -> last_score is not None
        trials.sort(key=lambda t: self._trial_state[t].last_score)

        if len(trials) <= 1:
            return [], []
        num_trials_in_quantile = math.ceil(len(trials) * self._quantile_fraction)
        if num_trials_in_quantile == len(trials) and self._quantile_fraction < 1.0:
            # have at least one upper
            num_trials_in_quantile -= 1
        if num_trials_in_quantile > len(trials) / 2:
            return trials[:num_trials_in_quantile], trials[num_trials_in_quantile:]
        return (trials[:num_trials_in_quantile], trials[-num_trials_in_quantile:])

    def _checkpoint_or_exploit(
        self,
        trial: Trial,
        tune_controller: "TuneController",
        upper_quantile: List[Trial],
        lower_quantile: List[Trial],
    ):
        """Checkpoint if in upper quantile, exploits if in lower."""
        state = self._trial_state[trial]
        # NEW: Create a checkpoint anyway
        logger.debug("Instructing %s to save.", trial)
        checkpoint = tune_controller._schedule_trial_save(trial, result=state.last_result)
        if trial in upper_quantile:
            # The trial last result is only updated after the scheduler
            # callback. So, we override with the current result.
            logger.debug("Trial %s is in upper quantile. Saving checkpoint.", trial)
            if trial.status == Trial.PAUSED:
                if trial.temporary_state.saving_to and isinstance(
                    trial.temporary_state.saving_to, _FutureTrainingResult
                ):
                    logger.debug("Trial %s is still saving.", trial)
                    state.last_checkpoint = trial.temporary_state.saving_to
                else:
                    # Paused trial will always have an in-memory checkpoint.
                    logger.debug("Trial %s is paused. Use last available checkpoint %s.", trial, trial.checkpoint)
                    state.last_checkpoint = trial.checkpoint
            else:
                logger.debug("Keeping checkpoint of trial %s for exploit.", trial)
                # TODO: # FIXME does this create two checkpoint with Trainable Auto saving?
                # state.last_checkpoint = tune_controller._schedule_trial_save(trial, result=state.last_result)
                state.last_checkpoint = checkpoint

            self._num_checkpoints += 1
        else:
            state.last_checkpoint = None  # not a top trial

        if trial in lower_quantile:
            trial_to_clone = random.choice(upper_quantile)  # if quantile is 0, isn't this empty?
            assert trial is not trial_to_clone
            clone_state = self._trial_state[trial_to_clone]
            last_checkpoint = clone_state.last_checkpoint

            logger.debug("Trial %s is in lower quantile. Exploiting trial %s.", trial, trial_to_clone)

            if isinstance(last_checkpoint, _FutureTrainingResult):
                training_result = last_checkpoint.resolve()

                if training_result:
                    clone_state.last_result = training_result.metrics
                    clone_state.last_checkpoint = training_result.checkpoint
                    last_checkpoint = clone_state.last_checkpoint
                else:
                    logger.debug(
                        "PBT-scheduled checkpoint save resolved to None. Trial "
                        "%s didn't save any checkpoint before "
                        "and can't be exploited.",
                        trial_to_clone,
                    )
                    last_checkpoint = None

            if not last_checkpoint:
                logger.info("[pbt]: no checkpoint for trial %s. Skip exploit for Trial %s", trial_to_clone, trial)
                return
            self._exploit(tune_controller, trial, trial_to_clone)

    def _check_result(self, result: dict):
        if self._time_attr not in result:
            time_missing_msg = (
                "Cannot find time_attr {} "
                "in trial result {}. Make sure that this "
                "attribute is returned in the "
                "results of your Trainable.".format(self._time_attr, result)
            )
            if self._require_attrs:
                raise RuntimeError(
                    time_missing_msg + "If this error is expected, you can change this to "
                    "a warning message by "
                    "setting PBT(require_attrs=False)"
                )
            if log_once("pbt-time_attr-error"):
                logger.warning(time_missing_msg)
        if self._metric not in result:
            metric_missing_msg = (
                "Cannot find metric {} in trial result {}. "
                "Make sure that this attribute is returned "
                "in the "
                "results of your Trainable.".format(self._metric, result)
            )
            if self._require_attrs:
                raise RuntimeError(
                    metric_missing_msg + "If this error is expected, "
                    "you can change this to a warning message by "
                    "setting PBT(require_attrs=False)"
                )
            if log_once("pbt-metric-error"):
                logger.warning(metric_missing_msg)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from jax import lax
from ray.rllib.core.models.base import ActorCriticEncoder
from ray.rllib.core.rl_module import RLModule
from typing_extensions import TypedDict

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import chex
    import jax.numpy as jnp
    from flax.training.train_state import TrainState
    from ray.rllib.utils.typing import StateDict

    from ray_utilities.jax.utils import ExtendedTrainState
    from ray_utilities.typing.model_return import ActorCriticEncoderOutput


class JaxActorCriticEncoder(ActorCriticEncoder):
    def __call__(self, inputs: dict, **kwargs) -> ActorCriticEncoderOutput[jnp.ndarray]:
        return self._forward(inputs, **kwargs)  # pyright: ignore[reportReturnType]  # interface untyped dict


# pyright: enableExperimentalFeatures=true
class JaxStateDict(TypedDict):
    module_key: int | chex.PRNGKey


class JaxActorCriticStateDict(JaxStateDict):
    actor: ExtendedTrainState | TrainState
    critic: TrainState
    module_key: int | chex.PRNGKey


class JaxModule(RLModule):
    """
    Attributes:
        states: A dictionary of state variables.

    Methods:
        - get_state
        - set_state
        - _forward_exploration calls self._forward with lax.stop_gradient(batch)
        - _forward_inference calls self._forward with lax.stop_gradient(batch)
    """

    def __init__(self, *args, **kwargs):
        # set before super; RLModule.__init__ will call setup
        self.states: JaxStateDict | StateDict | JaxActorCriticStateDict = {}
        super().__init__(*args, **kwargs)

    def get_state(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        *args,  # noqa: ARG002
        inference_only: bool = False,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ) -> JaxStateDict | StateDict | JaxActorCriticStateDict:
        # TODO: could return less than the full state if not inference_only; i.e. do not return the critic
        return self.states

    def set_state(self, state: JaxStateDict | StateDict | JaxActorCriticStateDict) -> None:
        # Note: Not entirely following Rllib interface
        if not state:
            logger.warning("State is empty, not setting state.")
        self.states = state

    def _forward_exploration(self, batch: dict[str, Any], **kwargs) -> dict[str, Any]:
        return self._forward(lax.stop_gradient(batch), **kwargs)

    def _forward_inference(self, batch: dict[str, Any], **kwargs) -> dict[str, Any]:
        return self._forward(lax.stop_gradient(batch), **kwargs)


if TYPE_CHECKING:
    JaxModule()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Generic, Mapping

from ray.rllib.core.models.base import ACTOR, CRITIC, ENCODER_OUT
from typing_extensions import NotRequired, TypeAliasType, TypedDict, TypeVar

if TYPE_CHECKING:
    import chex
    from ray.rllib.utils.typing import TensorType

logger = logging.getLogger(__name__)

_TensorT = TypeVar("_TensorT", bound="TensorType | chex.Array", default="TensorType")


# needed for extra_items
# pyright: enableExperimentalFeatures=true
class BatchDict(TypedDict, Generic[_TensorT], extra_items="_TensorT"):
    """Keys correspond to ray.rllib.core.columns.Columns"""

    action_dist_inputs: _TensorT
    action_logp: _TensorT


Batch = TypeAliasType("Batch", BatchDict[_TensorT] | Mapping[str, "_TensorT"], type_params=(_TensorT,))


class EncoderOut(TypedDict, Generic[_TensorT]):
    actor: Batch[_TensorT]
    critic: NotRequired[Batch[_TensorT]]


class ActorCriticEncoderOutput(TypedDict, Generic[_TensorT]):
    encoder_out: EncoderOut[_TensorT]


# Test implemented values

_bad_keys = []
if ENCODER_OUT not in ActorCriticEncoderOutput.__required_keys__:
    _bad_keys.append(ENCODER_OUT)
if ACTOR not in EncoderOut.__required_keys__:
    _bad_keys.append(ACTOR)
if CRITIC not in EncoderOut.__optional_keys__:
    _bad_keys.append(CRITIC)
if _bad_keys:
    logger.error("Keys %s have changed in RLlib; this module %s needs an update", _bad_keys, __name__)



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from ray.rllib.algorithms.algorithm import Algorithm

from ray_utilities.callbacks.algorithm.callback_mixins import BudgetMixin
from ray_utilities.callbacks.algorithm.dynamic_hyperparameter import DynamicHyperparameterCallback
from ray_utilities.dynamic_config.dynamic_buffer_update import get_dynamic_evaluation_intervals

if TYPE_CHECKING:
    from ray.rllib.utils.metrics.metrics_logger import MetricsLogger
    from ray.rllib.algorithms.algorithm import Algorithm


class AdjustConfigWhenBatchSizeChanges(BudgetMixin, DynamicHyperparameterCallback):
    def on_algorithm_init(self, *, algorithm: Algorithm, metrics_logger: MetricsLogger | None = None, **kwargs) -> None:
        super().on_algorithm_init(algorithm=algorithm, metrics_logger=metrics_logger, **kwargs)
        self._set_budget_on_algorithm_init(algorithm=algorithm, **kwargs)
        self._evaluation_intervals: dict[int, int] = dict(
            zip(
                self._budget["step_sizes"],
                (
                    get_dynamic_evaluation_intervals(
                        self._budget["step_sizes"],
                        eval_freq=algorithm.config.evaluation_interval,  # pyright: ignore[reportOptionalMemberAccess]
                        batch_size=algorithm.config.train_batch_size_per_learner,  # pyright: ignore[reportOptionalMemberAccess]
                        take_root=True,
                    )
                    # 0 for no evaluation
                    if algorithm.config.evaluation_interval  # pyright: ignore[reportOptionalMemberAccess]
                    else [0] * len(self._budget["step_sizes"])
                ),
            )
        )

    def adjust_config_based_on_batch_size(
        self,
        *,
        algorithm: "Algorithm",
        **kwargs,
    ) -> None:
        """
        on_algorithm_init callback

        Large / small batch sizes call for different config values:
        e.g.
        - smaller / larger evaluation_interval
        - broader / narrower smoothing window for metrics
        """
        # NOTE: Can interfere with DynamicEvalCallback when evaluation_interval changes
        self._update_algorithm(algorithm, key="evaluation_interval", value=...)
        # Cannot change window of metric logger when already logged
        self._update_algorithm(
            algorithm=algorithm,
            key="metrics_num_episodes_for_smoothing",
            value=max(30, min(300, int(math.sqrt(algorithm.config.train_batch_size_per_learner)))),  # pyright: ignore[reportOptionalMemberAccess]
        )

    def on_train_result(self, *, algorithm: Algorithm, metrics_logger: MetricsLogger | None = None, **kwargs) -> None:
        self.adjust_config_based_on_batch_size(algorithm=algorithm, **kwargs)



# pyright: enableExperimentalFeatures=true
from __future__ import annotations

import dataclasses
import logging
import math
from copy import deepcopy
from functools import partial
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Literal, Optional, TypeVar, cast, overload

import ray
from ray.experimental import tqdm_ray
from ray.rllib.algorithms import AlgorithmConfig
from ray.rllib.core import COMPONENT_LEARNER, COMPONENT_METRICS_LOGGER, COMPONENT_RL_MODULE
from ray.rllib.utils.metrics import (
    ALL_MODULES,  # pyright: ignore[reportPrivateImportUsage]
    ENV_RUNNER_RESULTS,
    EVALUATION_RESULTS,
    LEARNER_RESULTS,
    NUM_ENV_STEPS_SAMPLED_LIFETIME,
)
from ray.rllib.utils.metrics.metrics_logger import MetricsLogger
from typing_extensions import Sentinel, TypeAliasType

from ray_utilities.callbacks.algorithm.seeded_env_callback import SeedEnvsCallback
from ray_utilities.config import seed_environments_for_config
from ray_utilities.constants import NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME
from ray_utilities.dynamic_config.dynamic_buffer_update import calculate_iterations, calculate_steps
from ray_utilities.misc import AutoInt
from ray_utilities.warn import (
    warn_about_larger_minibatch_size,
    warn_if_batch_size_not_divisible,
    warn_if_minibatch_size_not_divisible,
)

if TYPE_CHECKING:
    from ray.rllib.algorithms import Algorithm
    from ray.rllib.callbacks.callbacks import RLlibCallback
    from ray.rllib.core.rl_module.default_model_config import DefaultModelConfig
    from ray.rllib.env.env_runner import EnvRunner
    from ray.rllib.env.multi_agent_env_runner import MultiAgentEnvRunner
    from ray.rllib.env.single_agent_env_runner import SingleAgentEnvRunner

    from ray_utilities.config import DefaultArgumentParser
    from ray_utilities.setup.experiment_base import AlgorithmType_co, ConfigType_co, ExperimentSetupBase, ParserType_co
    from ray_utilities.typing import (
        RewardUpdaters,
        StrictAlgorithmReturnData,
    )
    from ray_utilities.typing.algorithm_return import EvalEnvRunnersResultsDict
    from ray_utilities.typing.metrics import (
        AnyLogMetricsDict,
        LogMetricsDict,
        _LogMetricsEnvRunnersResultsDict,
        _LogMetricsEvalEnvRunnersResultsDict,
        _NewLogMetricsEvaluationResultsDict,
    )
    from ray_utilities.typing.trainable_return import RewardUpdater

logger = logging.getLogger(__name__)

DefaultExperimentSetup = TypeAliasType(
    "DefaultExperimentSetup", "ExperimentSetupBase[DefaultArgumentParser, AlgorithmConfig, Algorithm]"
)

_AlgorithmConfigT = TypeVar("_AlgorithmConfigT", bound="AlgorithmConfig")

_NOT_FOUND = Sentinel("_NOT_FOUND")


@overload
def episode_iterator(args: dict[str, Any], hparams: Any, *, use_pbar: Literal[False]) -> range: ...


@overload
def episode_iterator(args: dict[str, Any], hparams: dict[Any, Any], *, use_pbar: Literal[True]) -> tqdm_ray.tqdm: ...


def episode_iterator(args: dict[str, Any], hparams: dict[str, Any], *, use_pbar: bool = True) -> tqdm_ray.tqdm | range:
    """Creates an iterator for `args["iterations"]`

    Will create a `tqdm` if `use_pbar` is True, otherwise returns a range object.
    """
    if use_pbar:
        return tqdm_ray.tqdm(range(args["iterations"]), position=hparams.get("process_number", None))
    return range(args["iterations"])


def get_current_step(result: StrictAlgorithmReturnData | LogMetricsDict | dict[str, Any] | MetricsLogger) -> int:
    # requires exact_sampling_callback to be set in the results, otherwise fallback
    if isinstance(result, MetricsLogger):
        # For peek do not use None as default as this triggers KeyError
        current_step = result.peek((LEARNER_RESULTS, NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME), default=_NOT_FOUND)
        if current_step is not _NOT_FOUND:
            return current_step
        current_step = result.peek((ENV_RUNNER_RESULTS, NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME), default=_NOT_FOUND)
        if current_step is not _NOT_FOUND:
            return current_step
        return result.peek((ENV_RUNNER_RESULTS, NUM_ENV_STEPS_SAMPLED_LIFETIME), default=0)

    current_step = result.get("current_step")  # likely not present
    if current_step is not None:  # LogMetricsDict
        return current_step
    result = cast("StrictAlgorithmReturnData", result)
    if LEARNER_RESULTS in result:
        current_step = result[LEARNER_RESULTS][ALL_MODULES].get(NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME)
        if current_step is not None:
            return current_step
    # try metric logged on env runner; else defaults to NUM_ENV_STEPS_SAMPLED_LIFETIME
    return int(
        result[ENV_RUNNER_RESULTS].get(
            NUM_ENV_STEPS_PASSED_TO_LEARNER_LIFETIME, result[ENV_RUNNER_RESULTS][NUM_ENV_STEPS_SAMPLED_LIFETIME]
        )
    )


def patch_model_config(config: AlgorithmConfig, model_config: dict[str, Any] | DefaultModelConfig):
    """Updates the config.model_config and rl_module_spec.model_config with the given model_config."""
    if dataclasses.is_dataclass(model_config):
        model_config = dataclasses.asdict(model_config)
    model_config = deepcopy(model_config)
    if isinstance(config._model_config, dict):
        config._model_config.update(model_config)
    elif config._model_config is not None:
        config._model_config = dataclasses.asdict(config._model_config) | model_config
    else:
        config._model_config = model_config
    if config._rl_module_spec:
        if config._rl_module_spec.model_config is None:
            config._rl_module_spec.model_config = config._model_config
        elif isinstance(config._rl_module_spec.model_config, dict):
            config._rl_module_spec.model_config.update(model_config)
        else:
            config._rl_module_spec.model_config = dataclasses.asdict(config._rl_module_spec.model_config) | model_config


def _patch_with_param_space(
    args: dict[str, Any],
    config: _AlgorithmConfigT,
    *,
    hparams: dict[str, Any],
    config_inplace: bool = False,
) -> tuple[dict[str, Any], _AlgorithmConfigT]:
    same_keys = set(args.keys()) & set(hparams.keys())
    args["__overwritten_keys__"] = {}
    if "model_config" in hparams:
        patch_model_config(config, hparams["model_config"])
    if not same_keys:
        logger.debug("No keys to patch in args with hparams: %s", hparams)
        return args, config
    msg_dict = {k: f"{args[k]} -> {hparams[k]}" for k in same_keys}
    if (
        "train_batch_size_per_learner" in same_keys
        and config.minibatch_size is not None
        and config.minibatch_size > hparams["train_batch_size_per_learner"]
    ):
        warn_about_larger_minibatch_size(
            minibatch_size=config.minibatch_size,
            train_batch_size_per_learner=hparams["train_batch_size_per_learner"],
            note_adjustment=True,
        )
        msg_dict["minibatch_size"] = f"{config.minibatch_size} -> {hparams['train_batch_size_per_learner']}"
        if config_inplace:
            object.__setattr__(config, "minibatch_size", hparams["train_batch_size_per_learner"])
        args["__overwritten_keys__"]["minibatch_size"] = hparams["train_batch_size_per_learner"]

    logger.info("Patching args with hparams: %s", msg_dict)
    for key in same_keys:
        args[key] = hparams[key]
        if config_inplace:
            object.__setattr__(config, key, hparams[key])
        args["__overwritten_keys__"][key] = hparams[key]
    if not config_inplace:
        is_frozen = config._is_frozen
        config = cast("_AlgorithmConfigT", config.copy(copy_frozen=False))
        config.update_from_dict(args["__overwritten_keys__"])
        if is_frozen:
            config.freeze()
    assert config.minibatch_size or 0 <= config.train_batch_size_per_learner
    return args, config


def get_args_and_config(
    hparams: dict,
    setup: Optional["ExperimentSetupBase[Any, ConfigType_co, Any]"] = None,
    setup_class: Optional[type["ExperimentSetupBase[Any, ConfigType_co, Any]"]] = None,
    model_config: Optional[dict[str, Any] | DefaultModelConfig] = None,
) -> tuple[dict[str, Any], ConfigType_co]:
    """
    Constructs the args and config from the given hparams, setup or setup_class.
    Either `setup` or `setup_class` must be provided, if both are provided, `setup` will be used.

    This function can be used in a trainable during tuning.

    Args:
        hparams: The hyperparameters selected for the trial from the search space from ray tune.
            Should include an `cli_args` key with the parsed arguments if `setup` is not provided.
        setup: An instance of `DefaultExperimentSetup` that contains the configuration and arguments.
        setup_class: A class of `DefaultExperimentSetup` that can be used to create the configuration
            and arguments. Ignored if `setup` is provided.
        model_config: Optional dictionary of model configuration overrides to apply.

    Returns:
        A tuple containing the parsed args (as a dict) and an AlgorithmConfig.
        If `setup` is provided, the args will be a copy of `setup.args` created with `vars`.
    """
    # region setup config
    args: dict[str, Any]
    if setup:
        # TODO: Use hparams
        args = setup.args_to_dict()
        config: ConfigType_co = setup.config.copy(copy_frozen=False)  # pyright: ignore[reportAssignmentType]
    elif setup_class:
        args = deepcopy(hparams["cli_args"])
        # TODO: this should use the parameters from the search space
        config = setup_class.config_from_args(SimpleNamespace(**args))
    else:
        raise ValueError("Either setup or setup_class must be provided.")
    if model_config is not None:
        patch_model_config(config, model_config)
    args, config = _patch_with_param_space(args, config, hparams=hparams)
    # endregion

    # region seeding

    env_seed = hparams.get("env_seed", None)
    # Seeded environments - sequential seeds have to be set here, env_seed comes from Tuner
    if args["env_seeding_strategy"] == "sequential":
        # Warn if a seed is set but no env_seed is present
        if env_seed is None and "cli_args" in hparams and hparams["cli_args"]["seed"] is not None:
            logger.warning(
                "cli_args has a seed(%d) set but env_seed is None, sequential seeding will not work. "
                "Assure that env_seed is passed as a parameter when creating the Trainable, "
                "e.g. sample env_seed by tune. Falling back to cli_args seed and env_seeding_strategy='same'.",
                hparams["cli_args"]["seed"],
            )
            env_seed = hparams["cli_args"]["seed"]
        seed_environments_for_config(config, env_seed)
    elif args["env_seeding_strategy"] == "same":
        seed_environments_for_config(config, args["seed"])
    elif args["env_seeding_strategy"] == "constant":
        seed_environments_for_config(config, SeedEnvsCallback.env_seed)
    else:  # random
        seed_environments_for_config(config, None)

    # endregion seeding

    return args, config


def update_running_reward(new_reward: float, reward_array: list[float]) -> float:
    if not math.isnan(new_reward):
        reward_array.append(new_reward)
    running_reward = sum(reward_array[-100:]) / (min(100, len(reward_array)) or float("nan"))  # nan for 0
    return running_reward


def create_running_reward_updater(initial_array: Optional[list[float]] = None) -> RewardUpdater:
    """
    Creates a partial function that updates the running reward.

    The partial function is stateful in their reward_array, which is initialized as an empty list if
    `initial_array` is not provided.
    """
    return cast(
        "RewardUpdater", partial(update_running_reward, reward_array=initial_array if initial_array is not None else [])
    )


@overload
def setup_trainable(
    hparams: dict[str, Any],
    setup: Optional["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"] = None,
    setup_class: Optional[type["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"]] = None,
    config_overrides: Optional[ConfigType_co | dict[str, Any]] = None,
    model_config: Optional[dict[str, Any] | DefaultModelConfig] = None,
    *,
    create_algo: Literal[False],
) -> tuple[dict[str, Any], "ConfigType_co", None, "RewardUpdaters"]: ...


@overload
def setup_trainable(
    hparams: dict[str, Any],
    setup: Optional["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"] = None,
    setup_class: Optional[type["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"]] = None,
    config_overrides: Optional[ConfigType_co | dict[str, Any]] = None,
    model_config: Optional[dict[str, Any] | DefaultModelConfig] = None,
    *,
    create_algo: Literal[True] = True,
) -> tuple[dict[str, Any], "ConfigType_co", "AlgorithmType_co", "RewardUpdaters"]: ...


def setup_trainable(
    hparams: dict[str, Any],
    setup: Optional["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"] = None,
    setup_class: Optional[type["ExperimentSetupBase[ParserType_co, ConfigType_co, AlgorithmType_co]"]] = None,
    config_overrides: Optional[ConfigType_co | dict[str, Any]] = None,
    model_config: Optional[dict[str, Any] | DefaultModelConfig] = None,
    *,
    create_algo: bool = True,
) -> tuple[dict[str, Any], "ConfigType_co", "AlgorithmType_co | None", "RewardUpdaters"]:
    """
    Sets up the trainable by getting the args and config from the given hparams, setup or setup_class.
    Either `setup` or `setup_class` must be provided, if both are provided, `setup` will be used.

    Args:
        hparams: The hyperparameters selected for the trial from the search space from ray tune.
            Should include an `cli_args` key with the parsed arguments if `setup` is not provided.
        setup: An instance of `DefaultExperimentSetup` that contains the configuration and arguments.
        setup_class: A class of `DefaultExperimentSetup` that can be used to create the configuration
            and arguments. Ignored if `setup` is provided.
        create_algo: Will build or load from checkpoint when True (default). Pass False to skip this
            step, e.g. when the algorithm is created in a Trainable (and the one here is discarded).
        config_overrides: Optional dictionary of configuration overrides to apply.
        model_config: Optional dictionary of model configuration overrides to apply.

    Returns:
        A tuple containing the parsed args (as a dict), an AlgorithmConfig, and an Algorithm.

    Note:
        - The returned config of algorithm.config is frozen to prevent unexpected behavior this config.
        - The type of the Algorithm is determined by the `algo_class` attribute of the config.
            This is not entirely type-safe.

    See Also:
        :func:`get_args_and_config`: Helper function build the AlgorithmConfig.
    """
    args, config = get_args_and_config(
        hparams,
        setup=setup,
        setup_class=setup_class,
        model_config=model_config,
    )
    if config_overrides:
        if isinstance(config_overrides, AlgorithmConfig):
            config_overrides = config_overrides.to_dict()
        if "train_batch_size_per_learner" in config_overrides or "minibatch_size" in config_overrides:
            batch_size = config_overrides.get("train_batch_size_per_learner", config.train_batch_size_per_learner)
            minibatch_size = config_overrides.get("minibatch_size", config.minibatch_size)
            if minibatch_size > batch_size:
                warn_about_larger_minibatch_size(
                    minibatch_size=minibatch_size, train_batch_size_per_learner=batch_size, note_adjustment=True
                )
                config_overrides["minibatch_size"] = batch_size
        warn_if_batch_size_not_divisible(
            batch_size=config_overrides.get("train_batch_size_per_learner", config.train_batch_size_per_learner),
            num_envs_per_env_runner=config_overrides.get("num_envs_per_env_runner", config.num_envs_per_env_runner),
        )
        warn_if_minibatch_size_not_divisible(
            minibatch_size=config_overrides.get("minibatch_size", config.minibatch_size),
            num_envs_per_env_runner=config_overrides.get("num_envs_per_env_runner", config.num_envs_per_env_runner),
        )
        config = cast("ConfigType_co", config.update_from_dict(config_overrides))

    if "train_batch_size_per_learner" in hparams or (
        config_overrides and "train_batch_size_per_learner" in config_overrides
    ):
        # recalculate iterations and total_steps
        if isinstance(args["iterations"], AutoInt):
            new_iterations = calculate_iterations(
                dynamic_buffer=args["dynamic_buffer"],
                batch_size=config.train_batch_size_per_learner,
                total_steps=args["total_steps"],
                assure_even=not args["use_exact_total_steps"],
                min_size=args["min_step_size"],
                max_size=args["max_step_size"],
            )
            logger.info(
                "Adjusted iterations %d -> %d based on train_batch_size_per_learner %d",
                args["iterations"],
                new_iterations,
                config.train_batch_size_per_learner,
            )
            args["iterations"] = new_iterations
        else:
            logger.debug("Not adjusting iterations, as it was not an 'auto' value.")
    if create_algo:
        if not args["from_checkpoint"]:
            try:
                # new API; Note: copies config!
                algo = config.build_algo(use_copy=True)  # copy=True is default; maybe use False
            except AttributeError:
                algo = config.build()
            # FIXME too much info in model_config (most cli args)

        # Load from checkpoint
        elif checkpoint_loader := (
            setup or setup_class
        ):  # TODO: possibly do not load algorithm and let Trainable handle it (should be an option)
            algo = checkpoint_loader.algorithm_from_checkpoint(args["from_checkpoint"], config=config)
            sync_env_runner_states_after_reload(algo)
            if config.algo_class is not None and not isinstance(algo, config.algo_class):
                logger.warning(
                    "Loaded algorithm from checkpoint is not of the expected type %s, got %s. "
                    "Check your setup class %s.algo_class.",
                    config.algo_class,
                    type(algo),
                    type(setup) if setup is not None else setup_class,
                )
        else:
            # Should not happen, is covered by checks in get_args_and_config
            logger.warning("No setup or setup_class provided, using default PPOSetup. ")
            algo = cast("Algorithm", config.algo_class).from_checkpoint(args["from_checkpoint"])
    else:
        algo = None
    reward_updaters: RewardUpdaters = {
        "running_reward": create_running_reward_updater(),
        "eval_reward": create_running_reward_updater(),
        "disc_eval_reward": create_running_reward_updater(),
    }
    config.freeze()
    return (
        args,
        config,  # NOTE: a copy of algo.config
        algo,  # pyright: ignore[reportReturnType]
        reward_updaters,
    )


def get_total_steps(args: dict[str, Any], config: "AlgorithmConfig") -> int | None:
    return (
        args.get("total_steps", None)
        if args["iterations"] == "auto" or args["use_exact_total_steps"]
        else (
            calculate_steps(
                args["iterations"],
                total_steps_default=args["total_steps"],
                min_step_size=args["min_step_size"],
                max_step_size=args["max_step_size"],
            )
            if args["dynamic_buffer"]
            else (
                config.train_batch_size_per_learner
                * max(1, config.num_learners)  # pyright: ignore[reportArgumentType]
                * args["iterations"]
            )
        )
    )


def _set_env_runner_state(
    env_runner: EnvRunner | SingleAgentEnvRunner | MultiAgentEnvRunner,
    state_ref: ray.ObjectRef,
    config_ref: ray.ObjectRef,
):
    state: dict = ray.get(state_ref)
    if COMPONENT_METRICS_LOGGER not in state:
        raise KeyError(f"State dictionary missing required key '{COMPONENT_METRICS_LOGGER}'.")
    config: AlgorithmConfig = ray.get(config_ref)
    env_runner.config = config
    env_runner.metrics.set_state(state[COMPONENT_METRICS_LOGGER])
    if hasattr(env_runner, "num_envs"):
        if 0 != env_runner.num_envs != env_runner.config.num_envs_per_env_runner:  # pyright: ignore[reportAttributeAccessIssue]
            # local env runner can have zero envs when we are remote
            logger.error(
                "EnvRunner has %d envs, but config.num_envs_per_env_runner is %d. Recreating envs.",
                env_runner.num_envs,  # pyright: ignore[reportAttributeAccessIssue]
                env_runner.config.num_envs_per_env_runner,
            )
            env_runner.make_env()


def _clear_nan_stats(stat: dict[str, Any | list[list[float]]]):
    for k, v in stat.items():
        if k != "_hist":
            continue
        hist: list[list[float]] = v
        for i, h in enumerate(hist):
            hist[i] = [0.0 if (isinstance(x, float) and math.isnan(x)) else x for x in h]
    return stat


def split_sum_stats_over_env_runners(
    struct: Any, path: tuple[str, ...] = (), parent=None, *, num_env_runners: int
) -> Any:
    """
    As sum values are aggregated over all env runners, split them evenly over the env runners
    again for every to have roughly its own metric.

    Args:
        struct: The structure to split, can be a dict or a list.
        path: private, used to track the path in the structure.
        parent: private, used to track the parent structure.
        num_env_runners: The number of env runners to split the stats over.
    """
    if num_env_runners <= 1:
        return struct  # No need to split if only one env runner
    if isinstance(struct, dict):
        return {
            k: split_sum_stats_over_env_runners(v, (*path, k), struct, num_env_runners=num_env_runners)
            for k, v in struct.items()
        }
    if isinstance(struct, list):
        return [split_sum_stats_over_env_runners(v, path, parent, num_env_runners=num_env_runners) for v in struct]
    if parent is not None and parent["reduce"] == "sum" and parent["clear_on_reduce"] is False:
        if path[-1] == "values" or (path[-1] == "_hist" and parent["window"] in (None, float("inf"))):
            return struct / num_env_runners
    return struct


def nan_to_zero_hist_leaves(
    struct: Any,
    path: tuple[str, ...] = (),
    parent=None,
    *,
    key: Optional[str] = "_hist",
    remove_all: bool = False,
    replace: Any = 0.0,
) -> Any:
    """
    With a bug in ray updating a metric with -= value, where value could be NaN,
    replace such leafs with `replace`, 0 by default.

    Also useful for testing where nan != nan.
    """
    if isinstance(struct, dict):
        return {
            k: nan_to_zero_hist_leaves(v, (*path, k), struct, key=key, remove_all=remove_all, replace=replace)
            for k, v in struct.items()
        }
    if isinstance(struct, list):
        return [
            nan_to_zero_hist_leaves(v, path, parent, key=key, remove_all=remove_all, replace=replace) for v in struct
        ]
    if path and (key is None or path[-1] == key):
        # Only modify if parent has "reduce" == "sum"
        if remove_all or (
            parent is not None
            and parent["reduce"] == "sum"
            and parent["clear_on_reduce"] is False
            and parent["window"] in (None, float("inf"))
        ):
            return replace if (isinstance(struct, float) and math.isnan(struct)) else struct
    return struct


def sync_env_runner_states_after_reload(algorithm: Algorithm) -> None:
    """
    Syncs metric states for env runners, fixing a bug with restored metrics

    See my PR: https://github.com/ray-project/ray/pull/54148
    """
    assert algorithm.learner_group is not None
    assert algorithm.metrics
    assert algorithm.config
    assert algorithm.evaluation_config
    rl_module_state = algorithm.learner_group.get_state(
        components=COMPONENT_LEARNER + "/" + COMPONENT_RL_MODULE,
        inference_only=True,
    )[COMPONENT_LEARNER]

    # Sync states, especially env_steps

    metrics_state = algorithm.metrics.get_state()
    # State keys are "--" joined
    env_runner_metrics_state = {
        COMPONENT_METRICS_LOGGER: {
            "stats": {
                k.removeprefix(ENV_RUNNER_RESULTS + "--"): split_sum_stats_over_env_runners(
                    nan_to_zero_hist_leaves(v), num_env_runners=algorithm.config.num_env_runners or 1
                )
                for k, v in metrics_state["stats"].items()
                if k.startswith(ENV_RUNNER_RESULTS + "--")
            }
        }
    }
    # Do not sync EnvRunner seeds here as they are already set (or will be reset)
    env_runner_metrics_state[COMPONENT_METRICS_LOGGER]["stats"].pop("environments--seeds--seed_sequence", None)
    eval_stats = {
        k.removeprefix(EVALUATION_RESULTS + "--" + ENV_RUNNER_RESULTS + "--"): split_sum_stats_over_env_runners(
            nan_to_zero_hist_leaves(v), num_env_runners=algorithm.config.num_env_runners or 1
        )
        for k, v in metrics_state["stats"].items()
        if k.startswith("--".join((EVALUATION_RESULTS, ENV_RUNNER_RESULTS)))
    }
    eval_runner_metrics = {COMPONENT_METRICS_LOGGER: {"stats": eval_stats}}

    if algorithm.env_runner_group:
        algorithm.env_runner_group.sync_env_runner_states(
            config=algorithm.config,
            env_steps_sampled=algorithm.metrics.peek((ENV_RUNNER_RESULTS, NUM_ENV_STEPS_SAMPLED_LIFETIME), default=0),
            rl_module_state=rl_module_state,
            env_to_module=algorithm.env_to_module_connector,
            module_to_env=algorithm.module_to_env_connector,
            # env_runner_metrics=env_runner_metrics,
        )
        # Sync metrics
        state_ref = ray.put(env_runner_metrics_state)
        config_ref = ray.put(algorithm.config)

        algorithm.env_runner_group.foreach_env_runner(
            partial(_set_env_runner_state, state_ref=state_ref, config_ref=config_ref),
            remote_worker_ids=None,  # pyright: ignore[reportArgumentType]
            local_env_runner=True,
            # kwargs is not save to use here, not used on all code paths
            timeout_seconds=0.0,  # This is a state update -> Fire-and-forget.
        )

    if algorithm.eval_env_runner_group:  # XXX Why elif here in RLLib code?
        algorithm.eval_env_runner_group.sync_env_runner_states(
            config=algorithm.evaluation_config,
            # NOTE: Ray does not use EVALUATION_RESULTS here!
            env_steps_sampled=algorithm.metrics.peek(
                (EVALUATION_RESULTS, ENV_RUNNER_RESULTS, NUM_ENV_STEPS_SAMPLED_LIFETIME), default=0
            ),
            rl_module_state=rl_module_state,
            env_to_module=algorithm.env_to_module_connector,
            module_to_env=algorithm.module_to_env_connector,
            # env_runner_metrics=env_runner_metrics,
        )
        if eval_stats:
            eval_state_ref = ray.put(eval_runner_metrics)
            eval_config_ref = ray.put(algorithm.config.get_evaluation_config_object())
            algorithm.eval_env_runner_group.foreach_env_runner(
                partial(_set_env_runner_state, state_ref=eval_state_ref, config_ref=eval_config_ref),
                remote_worker_ids=None,  # pyright: ignore[reportArgumentType]
                local_env_runner=False,
                # kwargs is not save to use here, not used on all code paths
                timeout_seconds=0.0,  # This is a state update -> Fire-and-forget.
            )


def make_divisible(a: int, b: int | None) -> int:
    """Make a divisible by b"""
    if b is not None and a % b != 0:
        a = (a // b + 1) * b
    return a


def is_algorithm_callback_added(config: AlgorithmConfig, callback_class: type[RLlibCallback]) -> bool:
    return (
        config.callbacks_class is callback_class
        or (
            isinstance(callback_class, partial)
            and (
                config.callbacks_class is callback_class.func
                or (isinstance(config.callbacks_class, partial) and config.callbacks_class.func is callback_class.func)
            )
        )
        or (isinstance(config.callbacks_class, type) and issubclass(config.callbacks_class, callback_class))
        or (isinstance(config.callbacks_class, (list, tuple)) and callback_class in config.callbacks_class)
    )


def get_training_results(
    result: StrictAlgorithmReturnData | AnyLogMetricsDict | dict[str, Any],
) -> _LogMetricsEnvRunnersResultsDict:
    """
    Returns the training / env_runner results from the given result dict.

    This method is agnostic to RAY_UTILITIES_NEW_LOG_METRICS and works with both
    new and rllib metrics structure
    """
    if "training" in result:
        return result["training"]
    return result[ENV_RUNNER_RESULTS]


def get_evaluation_results(
    result: StrictAlgorithmReturnData | AnyLogMetricsDict | dict[str, Any],
) -> None | EvalEnvRunnersResultsDict | _LogMetricsEvalEnvRunnersResultsDict | _NewLogMetricsEvaluationResultsDict:
    if EVALUATION_RESULTS not in result:
        return None
    eval_results = result[EVALUATION_RESULTS]
    if ENV_RUNNER_RESULTS in eval_results:
        return eval_results[ENV_RUNNER_RESULTS]
    return eval_results



"""Callbacks for Ray RLlib algorithms and Ray Tune experiments.

This module contains callback classes for Ray RLlib algorithms and Ray Tune tuners,
plus utilities for cleaning command-line arguments before passing them to logging
callbacks.

Main Components:
    - Algorithm callbacks in :mod:`~ray_utilities.callbacks.algorithm`
    - Tuner callbacks in :mod:`~ray_utilities.callbacks.tuner`
    - :func:`remove_ignored_args`: Filter arguments for experiment logging
    - :data:`LOG_IGNORE_ARGS`: Arguments to exclude from logging
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Iterable, Mapping, TypeVar, cast, overload

if TYPE_CHECKING:
    import argparse

    from tap import Tap

__all__ = ["LOG_IGNORE_ARGS", "remove_ignored_args"]

_K = TypeVar("_K")
_V = TypeVar("_V")

LOG_IGNORE_ARGS = (
    "wandb",
    "comet",
    "not_parallel",
    "silent",
    "tags",  # handles separately by callbacks
    "use_comet_offline",
)
"""
Arguments that control logging and execution behavior rather than experiment hyperparameters.
These attributes of the `DefaultArgumentParser` or keys in the `Trainable`'s `config`
which should not processed by logging callbacks.
"""


@overload
def remove_ignored_args(  # pyright: ignore[reportOverlappingOverload]
    args: dict[_K, _V | Callable], *, remove: Iterable[_K | str] = LOG_IGNORE_ARGS
) -> dict[_K, _V]: ...


@overload
def remove_ignored_args(
    args: Tap | argparse.Namespace | Any, *, remove: Iterable[Any] = LOG_IGNORE_ARGS
) -> dict[str, Any]: ...


def remove_ignored_args(
    args: Mapping[_K, Any] | Tap | argparse.Namespace | Any, *, remove: Iterable[_K | str] = LOG_IGNORE_ARGS
) -> Mapping[_K, Any] | dict[str, Any]:
    """Remove logging-specific arguments from experiment parameters.

    Filters out arguments that control logging and execution behavior rather than
    experiment hyperparameters. Designed for cleaning parameters before passing
    to experiment tracking platforms.

    Args:
        args: Arguments to process (dict, :class:`argparse.Namespace`, :class:`tap.Tap`,
            or object with ``as_dict()`` method).
        remove: Argument names to remove. Defaults to :data:`LOG_IGNORE_ARGS`.

    Returns:
        Dictionary with specified arguments removed. Callable values are also filtered out.

    Example:
        >>> args = argparse.Namespace(lr=0.001, wandb=True, epochs=10)
        >>> clean_args = remove_ignored_args(args)
        >>> "wandb" in clean_args
        False
        >>> clean_args["lr"]
        0.001

    Note:
        The original argument object is not modified.
    """
    if not isinstance(args, (dict, Mapping)):
        if hasattr(args, "as_dict"):  # Tap
            args = args.as_dict()
        else:
            args = vars(args)
        args = cast("dict[str, Any]", args)
    return {k: v for k, v in args.items() if k not in remove and not callable(v)}



"""Enhanced logging utilities with colored output for better debugging experience.

This module provides the :func:`nice_logger` function to create loggers with colored,
formatted output that makes debugging Ray Tune and RLlib experiments more pleasant
and informative.

Example:
    Basic usage for setting up a colored logger::

        from ray_utilities.nice_logger import nice_logger

        logger = nice_logger(__name__, level="DEBUG")
        logger.info("This will be colored and nicely formatted")
        logger.warning("Warnings stand out with colors")
        logger.error("Errors are clearly visible")

See Also:
    :mod:`colorlog`: The underlying library used for colored logging
"""

from __future__ import annotations

import logging
from contextlib import contextmanager

import colorlog
from typing_extensions import Sentinel

# pyright: enableExperimentalFeatures=true
_NO_PACKAGE = Sentinel("_NO_PACKAGE")


def nice_logger(logger: logging.Logger | str, level: int | str | None = None) -> logging.Logger:
    """Create or modify a logger with colored formatting and enhanced readability.

    This function sets up a logger with colored output using :mod:`colorlog`, making
    it easier to distinguish between different log levels and trace debugging information
    during Ray Tune experiments and RLlib training.

    Args:
        logger: Either a :class:`logging.Logger` instance or a string name to create
            a new logger.
        level: The logging level to set. Can be an integer (from :mod:`logging` module)
            or a string like ``"DEBUG"``, ``"INFO"``, ``"WARNING"``, ``"ERROR"``.
            If ``None``, the current level is preserved.

    Returns:
        A configured :class:`logging.Logger` with colored formatting that includes:

        - Color-coded log levels for easy identification
        - File name, line number, and function name for debugging
        - Consistent formatting across all log messages

    Warning:
        If the logger already has handlers, a warning will be logged suggesting
        to remove them first to avoid duplicate log messages.

    Example:
        >>> logger = nice_logger("my_experiment", level="DEBUG")
        >>> logger.info("Starting Ray Tune experiment")
        [INFO][ my_file.py:42, main_function] : Starting Ray Tune experiment

        Using with an existing logger::

        >>> import logging
        >>> existing_logger = logging.getLogger("ray.rllib")
        >>> enhanced_logger = nice_logger(existing_logger, level="WARNING")

    Note:
        The colored formatter includes the following information:

        - **Log level** (colored based on severity)
        - **Filename and line number** where the log was called
        - **Function name** where the log was called
        - **Log message** content

        Colors help distinguish between different log levels:
        - DEBUG: Cyan
        - INFO: Green
        - WARNING: Yellow
        - ERROR: Red
        - CRITICAL: Bold red
    """
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    if isinstance(level, str):
        level = getattr(logging, level.upper())
    if level is not None:
        logger.setLevel(level)
    if logger.hasHandlers():
        logger.warning(
            "Making a richer logger, but logger %s already has handlers, consider removing them first.", logger
        )
    utilities_handler = colorlog.StreamHandler()
    formatter = colorlog.ColoredFormatter(
        "%(log_color)s[%(levelname)s][ %(filename)s:%(lineno)d, %(funcName)s] :%(reset)s %(message)s"
    )
    utilities_handler.setFormatter(formatter)
    logger.addHandler(utilities_handler)
    return logger


def set_project_log_level(
    pkg_logger: logging.Logger | str, level: int | str, pkg_name: str | None | _NO_PACKAGE = _NO_PACKAGE
) -> None:
    """
    Sets the logging level for a package logger and all its subloggers.

    Args:
        pkg_logger (logging.Logger | str): The package logger instance or its name as a string.
        level (int | str): The logging level to set (e.g., logging.INFO or "INFO").
        pkg_name (str | None | _NO_PACKAGE, optional): The package name. If None, only the given logger is affected.
            If _NO_PACKAGE, the package name is inferred from the logger's name.

    Returns:
        None

    Side Effects:
        - Changes the log level of the specified logger and all its handlers.
        - Recursively sets the log level for all child loggers belonging to the package.
        - Logs an info message when the log level is changed.
        - Logs an exception if a child logger's level cannot be set.

    Raises:
        None
    """
    # Set level on root package logger
    if isinstance(pkg_logger, str):
        pkg_logger = logging.getLogger(pkg_logger)
    if pkg_logger.getEffectiveLevel() not in (level, logging.getLevelName(level)):  # pyright: ignore[reportDeprecated]
        pkg_logger.info(
            "Changing log level of %s logger and all subloggers to %s",
            pkg_logger.name,
            logging.getLevelName(level) if isinstance(level, int) else level,
        )
    pkg_logger.setLevel(level)
    for h in pkg_logger.handlers:
        h.setLevel(level)
    if pkg_name is None:
        return

    if pkg_name is _NO_PACKAGE:
        pkg_name = pkg_logger.name.split(".")[0]

    # Also set level for any already-configured child loggers (skip placeholders)
    for name, lg in logging.Logger.manager.loggerDict.items():
        if not isinstance(lg, logging.Logger):
            continue
        if lg is pkg_logger:
            continue
        if name == pkg_name or name.startswith(pkg_name + "."):
            try:
                lg.setLevel(level)
                for h in lg.handlers:
                    h.setLevel(level)
            except Exception:  # noqa: BLE001
                lg.exception("Failed to set level with this logger")
                # be conservative: ignore any logger that can't be adjusted
                continue


@contextmanager
def change_log_level(logger: logging.Logger, new_level: logging._Level):
    """Context manager to temporarily change a logger's level."""
    old_level = logger.getEffectiveLevel()
    logger.setLevel(new_level)
    try:
        yield
    finally:
        logger.setLevel(old_level)
